import time

from autowsgr.fight.battle import BattlePlan
from autowsgr.fight.exercise import NormalExercisePlan
from autowsgr.fight.normal_fight import NormalFightPlan
from autowsgr.game.expedition import Expedition
from autowsgr.game.game_operation import get_rewards, repair_by_bath, set_support
from autowsgr.game.get_game_info import get_loot_and_ship, get_resources
from autowsgr.notification import miao_alert
from autowsgr.timer import Timer
from autowsgr.types import ConditionFlag


class DailyOperation:
    def __init__(self, timer: Timer, gap=300) -> None:
        self.timer = timer
        if timer.config.daily_automation is None:
            raise ValueError('未设置日常任务，请检查配置文件')
        self.config = timer.config.daily_automation
        self.gap = gap

        if self.config.auto_expedition:
            self.expedition_plan = Expedition(self.timer)

        if self.config.auto_battle:
            if not self.config.battle_type:
                raise ValueError('未设置战役类型，请检查配置文件')
            self.battle_plan = BattlePlan(
                self.timer,
                plan_path=self.config.battle_type,
            )
        if self.config.auto_exercise:
            self.exercise_plan = NormalExercisePlan(
                self.timer,
                plan_path='plan_1',
                fleet_id=self.config.exercise_fleet_id,
            )
            self.complete_time = None

        if self.config.auto_normal_fight:
            self.fight_plans: list[NormalFightPlan] = []
            self.fight_complete_times = []
            for plan in self.config.normal_fight_tasks if self.config.normal_fight_tasks else []:
                self.fight_plans.append(
                    NormalFightPlan(
                        self.timer,
                        plan_path=plan[0],
                        fleet_id=plan[1],
                    ),
                )
                self.fight_complete_times.append(
                    [0, plan[2], plan[0]],
                )  # 二元组， [已完成次数, 目标次数, 任务名称]

        self.start_time = self.last_time = time.time()

    def run(self):
        # 自动战役，直到超过次数
        if self.config.auto_battle:
            ret = ConditionFlag.OPERATION_SUCCESS
            cnt = 0
            while ret is not ConditionFlag.BATTLE_TIMES_EXCEED:
                ret = self.battle_plan.run()
                cnt += 1
                if cnt > 13:
                    self.timer.logger.warning('战役没能正常执行完毕, 请检查日志排查原因')
                    break

        # 自动开启战役支援
        if self.config.auto_set_support:
            set_support(self.timer, True)

        get_loot_and_ship(self.timer)  # 获取胖次掉落和船只掉落数据
        get_resources(self.timer)

        # 自动演习
        if self.config.auto_exercise:
            self.check_exercise()

        # 自动出征
        if self.config.auto_normal_fight:
            while self._has_unfinished() and self._ship_max() and self._loot_max():
                task_id = self._get_unfinished()

                plan = self.fight_plans[task_id]
                ret = plan.run()

                if ret == ConditionFlag.OPERATION_SUCCESS or ret == ConditionFlag.SL:
                    self.fight_complete_times[task_id][0] += 1
                elif ret == ConditionFlag.DOCK_FULL:
                    miao_alert.miao_alert(0)
                    break  # 不解装则结束出征

                if self.config.quick_repair_limit and self.timer.quick_repaired_cost >= int(
                    self.config.quick_repair_limit,
                ):
                    self.timer.logger.info(
                        f'快修消耗达到上限:{self.config.quick_repair_limit}，结束出征',
                    )
                    break

                if time.time() - self.last_time >= self.gap:
                    self._expedition()
                    self._gain_bonus()
                    if self.config.auto_exercise:
                        self.check_exercise()
                    self.last_time = time.time()

        # 自动远征
        while True:
            if self.config.auto_exercise:
                self.check_exercise()
            self._bath_repair()
            self._expedition()
            self._gain_bonus()
            time.sleep(self.gap)

    def _has_unfinished(self) -> bool:
        return any(times[0] < times[1] for times in self.fight_complete_times)

    def _get_unfinished_plan_description(self, i: int) -> str:
        plan_str = ''
        plan_str += f'正在执行的PLAN: {self.fight_complete_times[i][2]}, '
        plan_str += f'已出击次数/目标次数: {self.fight_complete_times[i][0]}/{self.fight_complete_times[i][1]}, '
        plan_str += f'消耗快修数量: {self.timer.quick_repaired_cost}, '
        plan_str += f'已掉落船数量: {self.timer.got_ship_num}'
        if self.timer.can_get_loot:
            plan_str += f', 已掉落胖次数量: {self.timer.got_loot_num}'
        return plan_str

    def _get_unfinished(self) -> int:
        for i, times in enumerate(self.fight_complete_times):
            if times[0] < times[1]:
                self.timer.logger.info(self._get_unfinished_plan_description(i))
                return i
        raise ValueError('没有未完成的任务')

    def _expedition(self) -> None:
        if self.config.auto_expedition:
            self.expedition_plan.run(True)

    def _gain_bonus(self) -> None:
        if self.config.auto_gain_bonus:
            get_rewards(self.timer)
            self.timer.go_main_page()

    def _bath_repair(self) -> None:
        if self.config.auto_bath_repair:
            repair_by_bath(self.timer)

    def _ship_max(self) -> bool:
        if not self.config.stop_max_ship:
            return True
        if self.timer.got_ship_num < 500:
            return True
        self.timer.logger.info('船只数量已达到上限，结束出征')
        return False

    def _loot_max(self) -> bool:
        if not self.config.stop_max_loot:
            return True
        if self.timer.got_loot_num < 50:
            return True
        self.timer.logger.info('胖次数量已达到上限，结束出征')
        return False

    def check_exercise(self) -> None:
        # 判断在哪个时间段
        now_time = time.localtime(time.time())
        hour = now_time.tm_hour
        if 0 <= hour < 12:
            self.new_time_period = '0:00-12:00'
        elif 12 <= hour < 18:
            self.new_time_period = '12:00-18:00'
        else:
            self.new_time_period = '18:00-23:59'

        if self.new_time_period != self.complete_time:
            self.timer.logger.info('即将执行演习任务')
            self.exercise_plan.run()
            self.complete_time = self.new_time_period
        else:
            self.timer.logger.debug('当前时间段演习已完成')



import os
import time

from autowsgr.fight.battle import BattleInfo, BattlePlan
from autowsgr.fight.common import ConditionFlag
from autowsgr.fight.event.event import Event
from autowsgr.game.expedition import Expedition
from autowsgr.game.game_operation import detect_ship_stats, quick_repair
from autowsgr.timer import Timer
from autowsgr.utils.io import yaml_to_dict


# image[1] 紧急状况

NODE_POSITION = {
    'main_to_event': (0.95, 0.5),
    'fight_to_event': (0.04, 0.06),
    'map_left': (59 / 960, 230 / 540),
    'map_right': (887 / 960, 230 / 540),
    'not_handle': (614 / 960, 435 / 540),
}


class EventFightPlan2025_0424(Event, BattlePlan):
    def __init__(
        self,
        timer: Timer,
        plan_path,
        fleet_id=None,
        event='20250424',
    ) -> None:
        """
        Args:
            fleet_id : 新的舰队参数, 优先级高于 plan 文件, 如果为 None 则使用计划参数.

        """
        if os.path.isabs(plan_path):
            plan_path = plan_path
        else:
            plan_path = timer.plan_tree['event'][event][plan_path]
        self.event_name = event
        BattlePlan.__init__(self, timer, plan_args=yaml_to_dict(plan_path))
        self.info = EventFightInfo2025_0424(timer, self.config.chapter, self.config.map)
        Event.__init__(self, timer, event)

    def _load_fight_info(self):
        self.info = EventFightInfo2025_0424(self.timer, self.config.chapter, self.config.map)

    def _go_fight_prepare_page(self) -> None:
        def recognize_map():
            return self.timer.wait_images(self.event_image[3:9]) + 1

        def change_fight_map(target_id):
            cur_id = recognize_map()
            while cur_id < target_id:
                self.timer.relative_click(*NODE_POSITION['map_right'])
                cur_id += 1
            while cur_id > target_id:
                self.timer.relative_click(*NODE_POSITION['map_left'])
                cur_id -= 1
            assert recognize_map() == target_id

        # 进入相应地图页面
        if not self.timer.image_exist(self.event_image[1]):  # 主页到地图
            self.timer.go_main_page()
            self.timer.relative_click(*NODE_POSITION['main_to_event'])

        change_fight_map(self.config.map)
        if not self.timer.click_image(self.event_image[1], timeout=10):
            self.timer.logger.warning('进入战斗准备页面失败,重新尝试进入战斗准备页面')
            self.timer.click_image(self.event_image[1], timeout=10)
        self.timer.relative_click(*NODE_POSITION['not_handle'])
        try:
            self.timer.wait_pages('fight_prepare_page', after_wait=0.15)
        except Exception as e:
            self.timer.logger.warning(
                f'匹配 fight_prepare_page 失败，尝试重新匹配, error: {e}',
            )
            self._go_fight_prepare_page()

    def _enter_fight(self) -> ConditionFlag:
        self._go_fight_prepare_page()
        self.info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.config.repair_mode, ship_stats=self.info.ship_stats)
        try:
            self.timer.relative_click(0.938, 0.926)  # 开始出征
            self.timer.wait_image(self.event_image[9])
            time.sleep(5)  # 等等吧
            self.timer.relative_click(0.938, 0.926, times=1)  # 点前进
            for _ in range(4):  # 多试几次
                if self.timer.image_exist(self.event_image[9]):
                    time.sleep(0.5)
                    self.timer.relative_click(0.938, 0.926, times=1)
            if not self.timer.image_exist(self.event_image[9]):  # 成功了
                return ConditionFlag.OPERATION_SUCCESS
        except TimeoutError:
            raise RuntimeError('进入战斗超时')

    def _make_decision(self, *args, **kwargs) -> ConditionFlag:
        state = self.update_state() if 'skip_update' not in kwargs else self.info.state
        if state == ConditionFlag.SL:
            return ConditionFlag.SL
        if self.info.state == 'goon_page':  # 估计是打完了
            time.sleep(3)
            for _ in range(8):
                time.sleep(0.5)
                self.timer.relative_click(*NODE_POSITION['fight_to_event'])
                value = self.timer.confirm_operation(False)
                if value:
                    return ConditionFlag.FIGHT_END
            raise RuntimeError('战斗结束确认失败')

        # 进行通用 NodeLevel 决策
        action, fight_stage = self.node.make_decision(
            self.info.state,
            self.info.last_state,
            self.info.last_action,
            self.info,
        )
        self.info.last_action = action
        return fight_stage

    def run_for_times(self, times, gap=1800) -> ConditionFlag:
        """多次执行同一任务, 自动进行远征操作
        Args:
            times (int): 任务执行总次数

            gap (int): 强制远征检查的间隔时间
        Raise:
            RuntimeError: 战斗进行时出现错误
        Returns:
            ConditionFlag
        """
        assert times >= 1
        expedition = Expedition(self.timer)
        self.timer.goto_game_page('map_page')
        for i in range(times):
            if time.time() - self.timer.last_expedition_check_time >= gap:
                self.timer.go_main_page()
                expedition.run(True)
            fight_flag = self.run()
            if fight_flag not in [ConditionFlag.SL, ConditionFlag.OPERATION_SUCCESS]:
                if fight_flag == ConditionFlag.DOCK_FULL:
                    return ConditionFlag.DOCK_FULL
                if fight_flag == ConditionFlag.SKIP_FIGHT:
                    return ConditionFlag.SKIP_FIGHT
                raise RuntimeError(f'战斗进行时出现异常, 信息为 {fight_flag}')
            self.timer.logger.info(f'已出击次数:{i+1}，目标次数{times}')
        return ConditionFlag.OPERATION_SUCCESS


class EventFightInfo2025_0424(Event, BattleInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20250424') -> None:
        self.chapter = chapter_id
        self.map = map_id
        BattleInfo.__init__(self, timer)
        Event.__init__(self, timer, event)
        self.end_page = 'unknown_page'
        self.successor_states['result'] = ['goon_page']
        self.state2image['goon_page'] = [self.event_image[9], 5]  # 打完后的标志



from autowsgr.fight.battle import BattlePlan
from autowsgr.fight.decisive_battle import DecisiveBattle
from autowsgr.fight.exercise import NormalExercisePlan
from autowsgr.fight.normal_fight import NormalFightPlan



import datetime
import queue
import threading
import time

from autowsgr.constants.custom_exceptions import ShipNotFoundErr
from autowsgr.fight.decisive_battle import DecisiveBattle, Logic
from autowsgr.game.build import BuildManager  # noqa: TC001
from autowsgr.game.expedition import Expedition
from autowsgr.game.game_operation import (
    change_ship,
    destroy_ship,
    detect_ship_stats,
    move_team,
    quick_repair,
)
from autowsgr.port.common import Ship
from autowsgr.port.ship import Fleet, count_ship
from autowsgr.timer.timer import Timer
from autowsgr.types import ConditionFlag, ShipType
from autowsgr.utils.api_image import crop_rectangle_relative
from autowsgr.utils.io import yaml_to_dict


def quick_register(timer: Timer, ships):
    def get_input():
        enable_quick = input() == 'y'
        q.put(enable_quick)

    print('检测到你在当前任务中启用了快速注册')
    print('请确认以下所有舰船均为绿血且不处于修复状态:')
    print(ships)
    print('确认吗(y/n)')

    q = queue.Queue()
    th = threading.Thread(target=get_input)
    th.start()
    th.join(10)
    if not q.empty() and q.get():
        for ship in ships:
            register_result = timer.port.register_ship(ship)
            if register_result is not None:
                register_result.level = 1
                register_result.status = 0
            else:
                timer.logger.error(f'注册舰船 {ship} 失败')
        return True
    timer.logger.info('放弃快速注册')
    return False


def register(timer: Timer, ships, fleet_id):
    if any(not timer.port.have_ship(ship) for ship in ships):
        timer.logger.info('含有未注册的舰船, 正在注册中...')

        # 检查当前舰队中是否含有未初始化的舰船并将其初始化
        timer.goto_game_page('fight_prepare_page')
        fleet = Fleet(timer, fleet_id)
        fleet.detect()
        if any(ship in ships for ship in fleet.ships):
            timer.logger.info('该舰队中处于修复状态的舰船正在被快速修复')
            quick_repair(timer, 3, switch_back=True)

        for i, ship in enumerate(fleet.ships):
            if ship in ships and not timer.port.have_ship(ship):
                register_result = timer.port.register_ship(ship)
                if register_result is not None:
                    level = fleet.levels[i]
                    if level is not None:
                        register_result.level = level
                        register_result.status = detect_ship_stats(timer)[i]
                    else:
                        timer.logger.error(f'未找到舰船 {ship} 的等级信息')
                else:
                    timer.logger.error(f'注册舰船 {ship} 失败')

        # 逐个初始化舰船, 效率较低, 待优化
        for ship in ships:
            if not timer.port.have_ship(ship):  #
                timer.logger.info(f'正在尝试注册 {ship}')
                try:
                    change_ship(timer, fleet_id, 1, ship)
                    quick_repair(timer, 3, switch_back=True)
                except ShipNotFoundErr:
                    timer.relative_click(0.05, 0.05)
                    timer.logger.warning(f'舰船 {ship} 注册失败, 放弃注册')
                    continue
                    # raise BaseException(f"未找到 {ship} 舰船")
                register_result = timer.port.register_ship(ship)
                if register_result is None:
                    timer.logger.error(f'注册舰船 {ship} 失败')
                    continue
                register_result.status = detect_ship_stats(timer)[1]
                fleet.detect()
                if fleet.levels[1] is None:
                    timer.logger.error(f'未找到舰船 {ship} 的等级信息')
                    continue
                register_result.level = fleet.levels[1]
        timer.port.show_fleet()


class Task:
    def __init__(self, timer: Timer) -> None:
        self.port = timer.port
        self.timer = timer

    def run(self):
        """执行任务
        Returns:
            bool: 任务是否结束, 如果任务正常结束, 则从 TaskRunner 中删除
            list:
        """


class FightTask(Task):
    def __init__(self, timer: Timer, file_path='', plan=None, *args, **kwargs) -> None:
        """
        Args:
            plan(FightPlan): 使用哪个战斗方案模板, 默认为 None.

            banned_ship (list(list(str))): 1-index 的列表, banned_ship[i] 表示第 i 号位不允许的舰船

            default_level_limit: 默认等级限制(2-111 之间的整数)

            level_limit (dict): 舰船名到等级限制的映射

            default_repair_mode: 默认修理方式

            repair_mode (dict): 舰船名到修理方式的映射

            ship_count: 舰队舰船数量

            fleet_id: 使用的舰队编号, 忽视 fight_plan 中的对应参数, 不支持第一舰队

            times: 任务重复次数

            max_repair_time: 最大修理时间: 超出该时间使用快修

            quick_repair: 无轮换时是否使用快修

            destroy_ship_types: 解装舰船

            all_ships: 所有参与轮换的舰船
        """
        super().__init__(timer)
        self.quick_register = False
        self.last_exec = time.time()
        self.plan = plan
        self.quick_repair = False
        self.destroy_ship_types = None
        self.default_level_limit = 2
        self.level_limit = {}
        self.default_repair_mode = 2
        self.repair_mode = {}
        self.ship_count = 0
        self.banned_ship = [[]] * 7
        self.fleet_id = 2
        self.times = 1
        self.all_ships = []
        if file_path != '':
            self.__dict__.update(yaml_to_dict(file_path))
        self.__dict__.update(kwargs)

        # 处理填写不规范的一些问题
        if self.repair_mode is None:
            self.repair_mode = {}
        if self.level_limit is None:
            self.level_limit = {}

        # 添加到舰船名字列表中
        for ship in self.all_ships:
            if ship not in self.timer.ship_names:
                self.timer.ship_names.append(ship)

        # 注册舰船 (等级直接设置成 1, 打完一遍后会更新掉)
        if self.quick_register and quick_register(timer, self.all_ships):
            return
        register(timer, self.all_ships, self.fleet_id)

    def build_fleet(self, ignore_statu=False):
        """尝试组建出征舰队"""

        def _build_fleet(ships):
            fleet = [
                None,
            ]
            ships = [ship for ship in ships if self.port.have_ship(ship)]
            ships = set(ships)
            for i in range(1, self.ship_count + 1):
                fleet.append(None)
                for ship in ships:
                    if self.banned_ship == -1 or ship not in self.banned_ship[i]:
                        fleet[i] = ship
                        ships.remove(ship)
                        break

            return None if any(ship is None for ship in fleet[1 : self.ship_count + 1]) else fleet

        if self.ship_count not in range(1, 7):
            raise ValueError(
                f'舰队舰船数量设置错误或者未设置, 当前设置的舰船数量为 {self.ship_count}',
            )
        # 清除已经达到等级要求的舰船并检查
        self.all_ships = [
            ship
            for ship in self.all_ships
            if self.port.get_ship_by_name(ship).level
            < self.level_limit.get(ship, self.default_level_limit)
        ]
        fleet = _build_fleet(self.all_ships)
        if fleet is None:
            self.timer.logger.info('由于等级超限无法组织舰队, 任务终止.')
            return 1, None

        # 清除不满足出征条件的舰船并检查
        fleet = _build_fleet(
            [
                ship
                for ship in self.all_ships
                if (
                    (
                        not ignore_statu
                        and self.port.get_ship_by_name(ship).status
                        < self.repair_mode.get(ship, self.default_repair_mode)
                    )
                    or ignore_statu
                )
            ],
        )
        if fleet is None:
            self.timer.logger.info(
                '由于部分舰船不满足出征条件而无法组织舰队, 任务暂停中...',
            )
            return 2, None
        return 0, fleet

    def check_repair(self):
        tasks = []
        for name in self.all_ships:
            ship = self.port.get_ship_by_name(name)
            if ship is None:
                self.timer.logger.warning(f'未找到舰船 {name}')
                continue
            if ship.status != 3 and ship.status >= self.repair_mode.get(
                ship,
                self.default_repair_mode,
            ):
                # 满足修理条件
                if ship.waiting_repair:
                    self.timer.logger.debug(
                        f'舰船 {name} 已在修理队列中, 不再重复添加.',
                    )
                else:
                    ship.waiting_repair = True
                    self.timer.logger.info(f'添加舰船 {name} 到修理队列.')
                    tasks.append(RepairTask(self.timer, ship))
        return tasks

    def run(self):
        # 应该退出的情况: 1.等级限制 2.不允许快修, 需要等待 3. 船坞已满, 需清理 4. 战斗任务已经执行完毕
        if self.times <= 0:
            return True, []

        status, fleet = self.build_fleet()
        if status == 1:
            return True, []
        if status == 2 and not self.quick_repair:
            return False, [*self.check_repair(), self]
        if self.port.ship_factory.full:
            tasks = [
                OtherTask(
                    self.timer,
                    'destroy',
                    destroy_ship_types=self.destroy_ship_types,
                ),
            ]
            return False, tasks
        if self.plan is None:
            raise ValueError('没有指定战斗策略')
        plan = self.plan
        plan.fleet = fleet
        plan.fleet_id = self.fleet_id
        plan.repair_mode = [3] * 6
        # 设置战时快修
        if status == 2:
            status, fleet = self.build_fleet(True)
            for i, name in enumerate(fleet):
                if name is None:
                    continue
                ship = self.port.get_ship_by_name(name)
                if ship is None:
                    self.timer.logger.warning(f'未找到舰船 {name}')
                    continue
                if ship.status >= self.repair_mode.get(name, self.default_repair_mode):
                    self.timer.logger.info(f'舰船 {name} 的状态已经标记为修复')
                    plan.repair_mode[i] = ship.status
                    ship.set_repair(0)
        # 执行战斗
        ret = plan.run()
        # 处理船坞已满
        if ret == ConditionFlag.DOCK_FULL:
            return False, [
                OtherTask(
                    self.timer,
                    'destroy',
                    destroy_ship_types=self.destroy_ship_types,
                ),
            ]
        self.times -= 1
        # 更新舰船状态
        if self.plan is None:
            self.timer.wait_pages('map_page')
        else:
            pass
        plan._go_fight_prepare_page()
        self.timer.goto_game_page('fight_prepare_page')
        move_team(self.timer, self.fleet_id)
        fleet = Fleet(self.timer)
        fleet.detect()
        ship_stats = detect_ship_stats(self.timer)
        for i, name in enumerate(fleet.ships):
            if name is None:
                continue
            ship = self.port.get_ship_by_name(name)
            level = fleet.levels[i]
            if ship is not None and level is not None:
                ship.level = level
                ship.status = ship_stats[i]

        return True, [*self.check_repair(), self]


class BuildTask(Task):
    def __init__(self, port) -> None:
        super().__init__(port)


class RepairTask(Task):
    def __init__(self, timer: Timer, ship: Ship, *args, **kwargs) -> None:
        super().__init__(timer)
        self.max_repiar_time = 1e9
        self.__dict__.update(kwargs)
        self.ship = ship

    def run(self):
        def switch_quick_repair(enable: bool):
            enabled = self.timer.check_pixel(
                (445, 91),
                (253, 150, 40),
                screen_shot=True,
            )
            if enabled != enable:
                self.timer.relative_click(0.464, 0.168)

        waiting = self.port.bathroom.get_waiting_time()
        if waiting is None:
            # 检查等待时间
            available_time = []
            timer = self.timer
            timer.goto_game_page('bath_page')
            baths = [(0.076, 0.2138), (0.076, 0.325), (0.076, 0.433)]
            repair_position = [
                (0.283, 0.544),
                (0.530, 0.537),
                (0.260, 0.656),
                (0.513, 0.644),
            ]

            for i in range(self.timer.config.bathroom_feature_count):
                timer.relative_click(*baths[i])
                for j in range(4):
                    timer.relative_click(*repair_position[j])
                    time.sleep(0.5)
                    timer.update_screen()
                    timer.logger.info(f'检查中:{(i, j)}')
                    if '快速修理' in [
                        result[1]
                        for result in self.recognize_screen_relative(
                            0.279,
                            0.319,
                            0.372,
                            0.373,
                        )
                    ]:
                        timer.logger.info('此位置有舰船修理中')
                        seconds = self.port.bathroom._time_to_seconds(
                            self.recognize_screen_relative(0.413, 0.544, 0.506, 0.596)[0][1],
                        )
                        timer.logger.info(f'预期用时: {seconds} 秒')
                        available_time.append(time.time() + seconds)
                        timer.relative_click(
                            797 / 1280,
                            509 / 720,
                            delay=1,
                        )  # 关闭快修选择界面

            while len(available_time) < timer.config.bathroom_count:
                available_time.append(0)
            info = str(
                [
                    datetime.datetime.fromtimestamp(timestamp).strftime(
                        '%Y-%m-%d %H:%M:%S',
                    )
                    for timestamp in available_time
                ],
            )
            self.port.bathroom.update_available_time(available_time)
            timer.logger.info(f'浴室信息检查完毕:{info}')

        # 扫描等待修理列表
        last_result = None
        self.timer.goto_game_page('choose_repair_page')
        while True:
            time_costs = self.recognize_screen_relative(
                0.041,
                0.866,
                0.966,
                0.943,
                True,
            )
            for time_cost in time_costs:
                self.timer.update_screen()
                text = time_cost[1].replace(' ', '')
                if text.startswith('耗时') and len(text) == 11:
                    # 整个图像截取完全
                    x = 0.041 + time_cost[0][0] / self.timer.screen.shape[1]
                    y = 0.741
                    img = crop_rectangle_relative(self.timer.screen, x - 0.06, y, 0.12, 0.042)
                    name = self.timer.recognize(img, candidates=self.timer.ship_names)
                    if len(name) == 0:
                        # 单字船名识别失败
                        continue
                    name = name[1]
                    seconds = self.port.bathroom._time_to_seconds(text[3:])
                    if name == self.ship.name:
                        if self.max_repiar_time <= seconds:
                            # 快速修复
                            self.timer.logger.info('满足快速修复条件, 使用快速修复工具')
                            self.ship.set_repair(0)
                            switch_quick_repair(True)
                            self.timer.relative_click(x, y)
                            return True, []

                        if not self.port.bathroom.is_available():
                            self.timer.logger.info(
                                '当前浴场已满, 不允许快速修复, 此任务延后',
                            )
                            return False, []

                        self.port.bathroom.add_repair(text[3:])
                        self.timer.relative_click(x, y)
                        self.timer.set_page('bath_page')
                        self.ship.set_repair(seconds)
                        return True, []

                    self.timer.logger.debug(f'识别到舰船: {name}')

            self.timer.relative_swipe(0.33, 0.5, 0.66, 0.5, delay=1)
            time.sleep(0.5)
            if time_costs == last_result:
                raise BaseException('未找到目标舰船')
            last_result = time_costs

    def recognize_screen_relative(
        self,
        left: float,
        top: float,
        right: float,
        bottom: float,
        update: bool = False,
    ):
        if update:
            self.timer.update_screen()
        return self.timer.recognize(
            crop_rectangle_relative(self.timer.screen, left, top, right - left, bottom - top),
            multiple=True,
        )


class OtherTask(Task):
    def __init__(self, timer: Timer, type, *args, **kwargs) -> None:
        """其它类型的任务
        Args:
            type (str): 任务类型
                "destroy": 舰船解装
                "empty": 不执行任何操作
                "build": 建造舰船
                "develop": 开发装备
            use_quick_build (bool): 是否允许使用快速修理
            resources_use (tuple): 按顺序油弹钢铝四项
        """
        super().__init__(timer)
        self.type = type
        if type == 'destroy':
            self.destroy_ship_types = kwargs['destroy_ship_types']
            for type in self.destroy_ship_types:
                if not isinstance(type, ShipType):
                    object.__setattr__(
                        self,
                        'destroy_ship_types',
                        [ShipType(t) for t in self.destroy_ship_types],
                    )
                    break
            timer.logger.info('船舱已满, 添加解装任务中...')
            if timer.port.ship_factory.waiting_destory:
                timer.logger.info('任务队列中已经有解装任务, 跳过')
                self.run = lambda self: None

        if type == 'build' or type == 'develop':
            try:
                self.resources_use = kwargs['resources_use']
                if 'use_quick_build' in kwargs:
                    self.use_quick_build = kwargs['use_quick_build']
                else:
                    timer.logger.warning('未指定是否使用快修, 默认不使用')
                    self.use_quick_build = False
            except:
                raise ValueError('未指定油弹钢铝数量')

    def run(self):
        if self.type == 'destroy':
            destroy_ship(self.timer, ship_types=self.destroy_ship_types)

        if self.type == 'build' or self.type == 'develop':
            factory: BuildManager = self.timer.port.factory
            type = 'ship' if self.type == 'build' else 'equipment'
            factory.update_slot_eta(type)
            if factory.has_empty_slot(type):
                factory.build(type, self.resources_use)
            elif not self.use_quick_build:
                factory.build(type, self.resources_use, True)
            else:
                return False, []
        return True, []


class DecisiveLogic(Logic):
    def __init__(self, timer, stats, level1, level2, flagship_priority) -> None:
        super().__init__(timer, stats, level1, level2, flagship_priority)

    def _get_best_fleet(self):
        return super().get_best_fleet()

    def get_best_fleet(self):
        def ship_available(ship) -> bool:
            # return ship in ships and self.timer.port.have_ship(ship) and self.timer.port.get_ship_by_name(ship).status < 2 # 大破修
            ship = self.timer.port.get_ship_by_name(ship)
            return (
                ship is not None
                and ship.name in ships
                and self.timer.port.have_ship(ship.name)
                and ship.status < 1
            )  # 中破修

        ships = self.stats.ships
        self.logger.debug(f'拥有舰船: {ships}')
        best_ships = [
            '',
        ]
        for ship in self.level1:
            if not ship_available(ship) or len(best_ships) == 7:
                continue
            best_ships.append(ship)
        for ship in self.level2:
            if not ship_available(ship) or len(best_ships) == 7 or ship in self.level1:
                continue
            best_ships.append(ship)

        for flag_ship in self.flag_ships:
            if flag_ship not in best_ships:
                continue
            p = best_ships.index(flag_ship)
            best_ships[p], best_ships[1] = best_ships[1], best_ships[p]
            break

        for _ in range(len(best_ships), 7):
            best_ships.append('')  # noqa: PERF401
        self.logger.debug(f'(考虑破损情况)当前最优：{best_ships}')
        return best_ships

    def _leave(self):
        return count_ship(self.get_best_fleet()) != count_ship(self._get_best_fleet())


class DecisiveFight(DecisiveBattle):
    rships: list

    def __init__(self, timer) -> None:
        super().__init__(timer)
        assert self.config is not None
        self.logic = DecisiveLogic(
            self.timer,
            self.stats,
            self.config.level1,
            self.config.level2,
            self.config.flagship_priority,
        )

    def can_fight(self):
        return count_ship(self.logic.get_best_fleet()) == count_ship(self.logic._get_best_fleet())


class DecisiveFightTask(Task):
    def __init__(self, timer: Timer, times, enable_quick_register=True, fleet_id=4) -> None:
        """
        Args:
            enable_quick_register (bool, optional): 是否启用快速注册. Defaults to False.
            fleet_id (int, optional): 注册舰队时占用的舰队编号. Defaults to 4.
        """
        super().__init__(timer)
        if timer.config.decisive_battle is None:
            raise ValueError('未配置决战任务')
        self.times = times

        if (
            timer.config.decisive_battle.level1 is None
            and timer.config.decisive_battle.level2 is None
        ):
            raise ValueError('未配置决战任务舰队')
        self.ships = timer.config.decisive_battle.level1 + timer.config.decisive_battle.level2

        self.db = DecisiveFight(self.timer)
        self.db.rships = self.ships
        if all(self.timer.port.have_ship(ship) for ship in self.ships):
            return
        if enable_quick_register and quick_register(timer, self.ships):
            return
        register(timer, self.ships, fleet_id)

    def check_repair(self):
        tasks = []
        for ship in self.ships:
            ship = self.timer.port.get_ship_by_name(ship)
            # if ship.status > 1 and ship.status < 3: # 大破修
            if ship.status > 0 and ship.status < 3:  # 中破修
                self.timer.logger.info(f'{ship.name} 需要修复')
                tasks.append(RepairTask(self.timer, ship))
        return tasks

    def run(self):
        while self.times > 0:
            if not self.db.can_fight():
                self.timer.logger.info('无法组合出足量的战斗舰船, 任务暂停中')
                return False, []

            res = self.db.run()

            if res == 'leave':
                self.timer.logger.info('战斗舰船发生破损, 暂离修复中....')
                return False, self.check_repair()
            self.times -= 1
            self.db = DecisiveFight(self.timer)
            self.db.rships = self.ships
        return True, []


class TaskRunner:
    def __init__(self, timer: Timer) -> None:
        self.tasks = []
        self.timer = timer

    def add_decisive_task(self, times=1):
        if any(isinstance(task, DecisiveFightTask) for task in self.tasks):
            raise ValueError('已经存在决战任务, 不允许再次添加')
        self.tasks.append(DecisiveFightTask(self.timer, times))

    def run(self):
        while True:
            # 调度逻辑: 依次尝试任务列表中的每个任务, 如果任务正常结束, 则从头开始, 否则找下一个任务.
            # 每个任务尝试执行后, 会向任务列表中添加一些新任务.
            id = 0
            while id < len(self.tasks):
                task = self.tasks[id]
                self.timer.logger.info(f'当前任务类型: {type(task)}')
                if 'times' in task.__dict__:
                    self.timer.logger.info(f'当前任务剩余次数: {task.times}')
                status, new_tasks = task.run()
                if status:
                    self.tasks = self.tasks[0:id] + new_tasks + self.tasks[id + 1 :]
                    break
                self.tasks = self.tasks[0 : id + 1] + new_tasks + self.tasks[id + 1 :]
                id += 1
            if len(self.tasks) == 0:
                self.timer.logger.info('本次全部任务已经执行完毕')
                return
            self.timer.logger.info('当轮任务已完成, 正在检查远征')
            time.sleep(30)
            Expedition(self.timer).run(True)



import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.fight.battle import BattleInfo, BattlePlan
from autowsgr.fight.common import start_march
from autowsgr.fight.event.event import PatrollingEvent
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.game.game_operation import detect_ship_stats, move_team, quick_repair
from autowsgr.timer import Timer


MAP_POSITIONS = [
    None,
    (100, 200),
    (500, 450),
    (520, 220),
    (350, 330),
    (600, 100),
    (700, 400),
]


class EventFightPlan20230117(PatrollingEvent, NormalFightPlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=None, event='20230117') -> None:
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)

    def _load_fight_info(self):
        # 地图文件位置: autowsgr/data/map/event/{event_name}
        # event_name 通常采用活动开始日期
        self.info = EventFightInfo20230117(self.timer, self.chapter, self.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'event', self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """对 normal_fight 的 _change_fight_map 的重写"""
        self.enter_map(chapter_id, map_id)
        while not self.timer.image_exist(
            self.common_image['little_monster'],
        ):  # 找到小怪物图标,点击下方进入主力决战
            while self.timer.wait_images_position(self.common_image['monster'], timeout=1) is None:
                self.random_walk()
            if self.timer.image_exist(self.common_image['little_monster']):
                break
            ret = self.timer.wait_images_position(self.common_image['monster'])
            self.timer.click(*ret)

        ret = self.get_close(self.common_image['little_monster'][0])
        self.timer.click(ret[0], ret[1] + 60)
        assert self.timer.wait_image(self.event_image[1])


class EventFightInfo20230117(PatrollingEvent, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20230117') -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)
        self.map_image = self.event_image[1]
        self.end_page = 'unknown_page'
        self.state2image['map_page'] = [self.map_image, 5]


class EventFightPlan20230117_2(PatrollingEvent, BattlePlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=None, event='20230117') -> None:
        self.event_name = event
        self.fleet_id = fleet_id
        BattlePlan.__init__(self, timer, plan_path)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)

    def _enter_fight(self):
        self._go_map_page()
        self.enter_map(self.chapter, self.map)
        assert self.timer.wait_image(self.event_image[2])
        self.find(self.enemy_image[self.target])
        ret = self.get_close(self.enemy_image[self.target])
        self.timer.click(*ret)
        assert self.timer.wait_image(self.event_image[4])
        self.timer.click(900, 500)
        self.timer.wait_pages('fight_prepare_page', after_wait=0.15)
        move_team(self.timer, self.fleet_id)
        self.info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.repair_mode, self.info.ship_stats)
        return start_march(self.timer)


class EventFightInfo20230117_2(PatrollingEvent, BattleInfo):
    def __init__(self, timer: Timer, chapter, map, event='20230117') -> None:
        BattleInfo.__init__(self, timer, chapter, map)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)
        self.end_page = 'unknown_page'
        self.state2image['battle_page'] = [self.event_image[2], 5]



from os.path import dirname, join


DATA_ROOT = join(dirname(dirname(__file__)), 'data')
IMG_ROOT = join(DATA_ROOT, 'images')
MAP_ROOT = join(DATA_ROOT, 'map')
SETTING_ROOT = join(DATA_ROOT, 'settings')
OCR_ROOT = join(DATA_ROOT, 'ocr')

BIN_ROOT = join(dirname(DATA_ROOT), 'bin')



import os
import sys
from enum import Enum

from autowsgr.utils.operator import unzip_element


class BaseEnum(Enum):
    """提供更友好的中文报错信息"""

    @classmethod
    def _missing_(cls, value: str) -> None:
        supported_values = ', '.join(
            [enum_object.value.__str__() for enum_object in cls.__members__.values()],
        )
        raise ValueError(f'"{value}" 不是合法的{cls.__name__}取值. 支持的有: [{supported_values}]')

    @classmethod
    def enum(cls) -> list:
        return list(cls.__members__.values())


class StrEnum(str, BaseEnum):
    @classmethod
    def get_all_chars(cls) -> list:
        char_list = []
        for name in unzip_element(cls.enum()):
            for char in name:
                char_list = list({*char_list, char})
        return char_list


class IntEnum(int, BaseEnum):
    pass


"""如果有一些功能在主程序中尚未支持（比如linux系统），请在本模块中对其进行raise
   主程序中将不考虑对这些异常的处理
"""


class OcrBackend(StrEnum):
    easyocr = 'easyocr'
    paddleocr = 'paddleocr'


class OSType(StrEnum):
    windows = 'Windows'
    linux = 'Linux'
    macos = 'macOS'

    @classmethod
    def auto(cls) -> 'OSType':
        if sys.platform.startswith('win'):
            return OSType.windows
        if sys.platform == 'darwin':
            return OSType.macos
        raise ValueError(f'不支持的操作系统 {sys.platform}')


class EmulatorType(StrEnum):
    leidian = '雷电'
    bluestacks = '蓝叠'
    mumu = 'MuMu'
    yunshouji = '云手机'
    others = '其他'

    def default_emulator_name(self, os: OSType) -> str:
        """自动获取默认模拟器连接名称"""
        if os == OSType.windows:
            match self:
                case EmulatorType.leidian:
                    return 'emulator-5554'
                case EmulatorType.bluestacks:
                    return '127.0.0.1:5555'
                case EmulatorType.mumu:
                    return '127.0.0.1:16384'
                case _:
                    raise ValueError(f'没有为 {self.value} 模拟器设置默认emulator_name，请手动指定')
        elif os == OSType.macos:
            match self:
                case EmulatorType.bluestacks:
                    return '127.0.0.1:5555'
                case EmulatorType.mumu:
                    return '127.0.0.1:5555'
                case _:
                    raise ValueError(f'没有为 {self.value} 模拟器设置默认emulator_name，请手动指定')
        else:
            raise ValueError(f'没有为 {os} 操作系统设置默认emulator_name，请手动指定')

    def auto_emulator_path(self, os: OSType) -> str:
        """自动获取模拟器路径"""
        adapter_fun = {
            OSType.windows: self.windows_auto_emulator_path,
            OSType.macos: self.macos_auto_emulator_path,
        }
        if os in adapter_fun:
            return adapter_fun[os]()
        raise ValueError(f'没有为 {os} 操作系统设置emulator_path查找方法，请手动指定')

    def windows_auto_emulator_path(self) -> str:
        """Windows自动识别模拟器路径"""
        import winreg

        try:
            match self:
                case EmulatorType.leidian:
                    with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\leidian') as key:
                        sub_key = winreg.EnumKey(key, 0)
                        with winreg.OpenKey(key, sub_key) as sub_key:
                            path, _ = winreg.QueryValueEx(sub_key, 'InstallDir')
                            return os.path.join(path, 'dnplayer.exe')
                case EmulatorType.bluestacks:
                    with winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE,
                        r'SOFTWARE\BlueStacks_nxt_cn',
                    ) as key:
                        path, _ = winreg.QueryValueEx(key, 'InstallDir')
                        return os.path.join(path, 'HD-Player.exe')
                case EmulatorType.mumu:
                    with winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE,
                        r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\MuMuPlayer-12.0',
                    ) as key:
                        path, _ = winreg.QueryValueEx(key, 'UninstallString')
                        return os.path.join(os.path.dirname(path), 'shell', 'MuMuPlayer.exe')
                case _:
                    raise ValueError(f'没有为 {self.value} 设置安装路径查找方法，请手动指定')
        except FileNotFoundError:
            raise FileNotFoundError(f'没有找到 {self.value} 的安装路径')

    def macos_auto_emulator_path(self) -> str:
        """macOS自动识别模拟器路径"""
        match self:
            case EmulatorType.mumu:
                path = '/Applications/MuMuPlayer.app'
            case EmulatorType.bluestacks:
                path = '/Applications/BlueStacks.app'
            case _:
                raise ValueError(f'没有为 {self.value} 设置安装路径查找方法，请手动指定')

        if os.path.exists(path):
            return path
        if os.path.exists(f'~/{path}'):
            # 全局安装目录-不存在的时候再去当前用户应用目录
            return f'~/{path}'
        raise FileNotFoundError(f'没有找到 {self.value} 的安装路径')


class GameAPP(StrEnum):
    official = '官服'
    xiaomi = '小米'
    tencent = '应用宝'

    @property
    def app_name(self) -> str:
        match self:
            case GameAPP.official:
                return 'com.huanmeng.zhanjian2'
            case GameAPP.xiaomi:
                return 'com.hoolai.zjsnr.mi'
            case GameAPP.tencent:
                return 'com.tencent.tmgp.zhanjian2'
            case _:
                raise ValueError(f'没有为 {self} 设置包名，请手动指定')


class RepairMode(IntEnum):
    moderate_damage = 1
    """中破就修"""
    severe_damage = 2
    """大破才修"""
    repairing = 3
    """正在修理中的"""


class FightCondition(IntEnum):
    steady_advance = 1
    """稳步前进"""
    firepower_forever = 2
    """火力万岁"""
    caution = 3
    """小心翼翼"""
    aim = 4
    """瞄准"""
    search_formation = 5
    """搜索阵型"""

    @property
    def relative_click_position(self) -> tuple[float, float]:
        return [
            None,
            [0.215, 0.409],
            [0.461, 0.531],
            [0.783, 0.362],
            [0.198, 0.764],
            [0.763, 0.74],
        ][self.value]


class Formation(IntEnum):
    single_column = 1
    """单纵阵"""
    double_column = 2
    """复纵阵"""
    circular = 3
    """轮型阵"""
    wedge = 4
    """梯形阵"""
    single_horizontal = 5
    """单横阵"""

    @property
    def relative_position(self) -> tuple[float, float]:
        return 0.597, self.value * 0.185 - 0.037


class SearchEnemyAction(StrEnum):
    no_action = 'no_action'
    retreat = 'retreat'
    detour = 'detour'
    refresh = 'refresh'


class ShipType(StrEnum):
    CV = '航母'
    CVL = '轻母'
    AV = '装母'
    BB = '战列'
    BBV = '航战'
    BC = '战巡'
    CA = '重巡'
    CAV = '航巡'
    CLT = '雷巡'
    CL = '轻巡'
    BM = '重炮'
    DD = '驱逐'
    SSG = '导潜'
    SS = '潜艇'
    SC = '炮潜'
    NAP = '补给'
    ASDG = '导驱'
    AADG = '防驱'
    KP = '导巡'
    CG = '防巡'
    CBG = '大巡'
    BG = '导战'
    Other = '其他'

    @property
    def relative_position_in_destroy(self) -> tuple[float, float]:
        dict = {
            ShipType.CV: (0.555, 0.197),
            ShipType.CVL: (0.646, 0.197),
            ShipType.AV: (0.738, 0.197),
            ShipType.BB: (0.830, 0.197),
            ShipType.BBV: (0.922, 0.197),
            ShipType.BC: (0.556, 0.288),
            ShipType.CA: (0.646, 0.288),
            ShipType.CAV: (0.738, 0.288),
            ShipType.CLT: (0.830, 0.288),
            ShipType.CL: (0.922, 0.288),
            ShipType.BM: (0.556, 0.379),
            ShipType.DD: (0.646, 0.379),
            ShipType.SSG: (0.738, 0.379),
            ShipType.SS: (0.830, 0.379),
            ShipType.SC: (0.922, 0.379),
            ShipType.NAP: (0.555, 0.470),
            ShipType.ASDG: (0.646, 0.470),
            ShipType.AADG: (0.738, 0.470),
            ShipType.KP: (0.830, 0.470),
            ShipType.CG: (0.922, 0.470),
            ShipType.CBG: (0.555, 0.561),
            ShipType.BG: (0.646, 0.561),
            ShipType.Other: (0.738, 0.561),
        }
        return dict[self]


class DestroyShipWorkMode(IntEnum):
    """拆解工作模式"""

    disable = 0
    """不启用舰种分类"""
    include = 1
    """拆哪些船"""
    exclude = 2
    """不拆哪些船"""


class ConditionFlag(StrEnum):
    DOCK_FULL = 'dock is full'
    """船坞已满并且没有设置自动解装"""
    FIGHT_END = 'fight end'
    """战斗结束标志, 一般不返回这个, 和 success 相同"""
    FIGHT_CONTINUE = 'fight continue'
    """战斗继续"""
    OPERATION_SUCCESS = 'success'
    """战斗流程正常结束 (到达了某个结束点或者选择了回港)"""
    BATTLE_TIMES_EXCEED = 'out of times'
    """战斗超时"""
    SKIP_FIGHT = 'skip fight'
    """跳过战斗"""
    SL = 'SL'
    """需要 / 进行了 SL 操作"""


class FormationName(StrEnum):
    single_column = '单纵'
    double_column = '复纵'
    circular = '轮型'
    wedge = '梯形'
    single_horizontal = '单横'



import copy
import os
import time

from autowsgr.configs import FightConfig
from autowsgr.constants.custom_exceptions import ImageNotFoundErr
from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.fight.common import DecisionBlock, FightInfo, FightPlan, start_march
from autowsgr.game.game_operation import change_ships, move_team, quick_repair
from autowsgr.game.get_game_info import detect_ship_stats, get_enemy_condition
from autowsgr.timer import Timer
from autowsgr.types import ConditionFlag
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict
from autowsgr.utils.math_functions import cal_dis


"""
常规战决策模块/地图战斗用模板
"""

MAP_NUM = [5, 6, 4, 4, 5, 4, 5, 5, 4]  # 每一章的地图数量


class NormalFightInfo(FightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id) -> None:
        super().__init__(timer)

        self.point_positions = None
        self.map_image = IMG.identify_images.map_page
        self.ship_image = [IMG.symbol_image[8], IMG.symbol_image[13]]
        self.chapter = chapter_id  # 章节名,战役为 'battle', 演习为 'exercise'
        self.map = map_id  # 节点名
        self.ship_position = (0, 0)
        self.node = 'A'  # 常规地图战斗中,当前战斗点位的编号
        # 实现通用 FightInfo 接口
        self.end_page = 'map_page'
        self.successor_states = {
            'proceed': {
                'yes': [
                    'fight_condition',
                    'spot_enemy_success',
                    'formation',
                    'fight_period',
                    'map_page',
                ],
                'no': ['map_page'],
            },
            'fight_condition': ['spot_enemy_success', 'formation', 'fight_period'],
            'spot_enemy_success': {
                'detour': [
                    'fight_condition',
                    'spot_enemy_success',
                    'formation',
                    'fight_period',
                ],
                'retreat': ['map_page'],
                'fight': ['formation', 'fight_period'],
            },
            'formation': ['fight_period'],
            'fight_period': ['night', 'result'],
            'night': {
                'yes': ['result'],
                'no': [['result', 10]],
            },
            'result': [
                'proceed',
                'map_page',
                'get_ship',
                'flagship_severe_damage',
            ],  # 两页战果
            'get_ship': ['proceed', 'map_page', 'flagship_severe_damage'],  # 捞到舰船
            'flagship_severe_damage': ['map_page'],
        }

        self.state2image = {
            'proceed': [IMG.fight_image[5], 7.5],
            'fight_condition': [IMG.fight_image[10], 22.5],
            'spot_enemy_success': [IMG.fight_image[2], 22.5],
            'formation': [IMG.fight_image[1], 22.5],
            'fight_period': [IMG.symbol_image[4], 30],
            'night': [IMG.fight_image[6], 150],
            'result': [IMG.fight_image[3], 90],
            'get_ship': [self.ship_image, 5],
            'flagship_severe_damage': [IMG.fight_image[4], 7.5],
            'map_page': [self.map_image, 7.5],
        }

        self.after_match_delay = {
            'night': 1.75,
            'proceed': 0.5,
            'get_ship': 1,
        }

    def reset(self):
        self.fight_history.reset()
        self.last_state = 'proceed'
        self.last_action = 'yes'
        self.state = 'proceed'  # 初始状态等同于 proceed 选择 yes

    def _before_match(self):
        # 点击加速
        if self.last_state in ['proceed', 'fight_condition'] or self.last_action == 'detour':
            self.timer.click(250, 520, delay=0, enable_subprocess=True)

        self.timer.update_screen()

        # 在地图上走的过程中获取舰船位置
        if self.last_state == 'proceed' or self.last_action == 'detour':
            self._update_ship_position()
            self._update_ship_point()
            pos = self.timer.get_image_position(IMG.confirm_image[3], False, 0.8)
            if pos:
                self.timer.confirm_operation(delay=0.25, must_confirm=1, confidence=0.8)

    def _after_match(self):
        # 在某些State下可以记录额外信息
        if self.state == 'spot_enemy_success':
            get_enemy_condition(self.timer, 'fight')
        super()._after_match()

    # ======================== Functions ========================

    def _update_ship_position(self):
        """在战斗移动界面(有一个黄色小船在地图上跑)更新黄色小船的位置"""
        pos = self.timer.get_image_position(IMG.fight_image[7], False, 0.8)
        if pos is None:
            pos = self.timer.get_image_position(IMG.fight_image[8], False, 0.8)
        if pos is None:
            return
        self.ship_position = pos
        if self.timer.config.show_map_node:
            self.timer.logger.debug(self.ship_position)

    def _update_ship_point(self):
        """更新黄色小船(战斗地图上那个)所在的点位 (1-1A 这种,'A' 就是点位)"""

        self.node = 'A'
        for i in range(26):
            ch = chr(ord('A') + i)
            if ch not in self.point_positions:
                continue
            if cal_dis(self.point_positions[ch], self.ship_position) < cal_dis(
                self.point_positions[self.node],
                self.ship_position,
            ):
                self.node = ch

        if self.timer.config.show_map_node:
            self.timer.logger.debug(self.node)

    def load_point_positions(self, map_path):
        """地图文件命名格式: [chapter]-[map].yaml"""
        self.point_positions = yaml_to_dict(
            os.path.join(map_path, str(self.chapter) + '-' + str(self.map) + '.yaml'),
        )


def check_blood(blood, rule) -> bool:
    """检查血量状态是否满足前进条件
        >>>check_blood([None, 1, 1, 1, 2, -1, -1], [2, 2, 2, -1, -1, -1])

        >>>True
    Args:
        blood (list): 1-based
        rule (list): 0-based
    """
    for i in range(max(len(blood) - 1, len(rule))):
        if blood[i + 1] == -1 or rule[i] == -1:
            continue
        if blood[i + 1] >= rule[i]:
            return False
    return True


class NormalFightPlan(FightPlan):
    """ " 常规战斗的决策模块"""

    """ 多点战斗基本模板 """

    def __init__(self, timer: Timer, plan_path, fleet_id=None, fleet=-1) -> None:
        """初始化决策模块,可以重新指定默认参数,优先级更高

        Args:
            fleet_id: 指定舰队编号, 如果为 None 则使用计划中的参数

            plan_path: 绝对路径 / 以 PLAN_ROOT 为根的相对路径

            fleet: 舰队成员, ["", "1号位", "2号位", ...], 如果为 None 则全部不变, 为 "" 则该位置无舰船, 为 -1 则不覆盖 yaml 文件中的参数

        Raises:
            BaseException: _description_
        """
        super().__init__(timer)
        # 从配置文件加载计划
        plan_path = (
            plan_path
            if os.path.isabs(plan_path)
            else self.timer.plan_tree['normal_fight'][plan_path]
        )
        plan_args = yaml_to_dict(plan_path)
        if fleet_id is not None:
            plan_args['fleet_id'] = fleet_id  # 舰队编号
        if fleet != -1:
            plan_args['fleet'] = fleet

        self.config = FightConfig.from_dict(plan_args)

        # 加载节点配置
        self.nodes: dict[str, DecisionBlock] = {}
        for node_name in self.config.selected_nodes:
            node_args = copy.deepcopy(plan_args.get('node_defaults', {}))
            if (
                'node_args' in plan_args
                and plan_args['node_args'] is not None
                and node_name in plan_args['node_args']
            ):
                node_args = recursive_dict_update(
                    node_args,
                    plan_args['node_args'][node_name],
                )
            self.nodes[node_name] = DecisionBlock(timer, node_args)
        self._load_fight_info()

    def _load_fight_info(self) -> None:
        # 信息记录器
        self.info = NormalFightInfo(self.timer, self.config.chapter, self.config.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'normal'))

    def _go_map_page(self) -> None:
        """活动多点战斗必须重写该模块"""
        """ 进入选择战斗地图的页面 """
        """ 这个模块负责的工作在战斗结束后如果需要进行重复战斗, 则不会进行 """
        self.timer.goto_game_page('map_page')

    def _go_fight_prepare_page(self) -> None:
        """活动多点战斗必须重写该模块"""
        """(从当前战斗结束后跳转到的页面)进入准备战斗的页面"""
        if not self.timer.ui.is_normal_fight_prepare:
            self.timer.goto_game_page('map_page')
        self.timer.goto_game_page('fight_prepare_page')
        self.timer.ui.is_normal_fight_prepare = True

    def _enter_fight(self, *args, **kwargs):
        """
        从任意界面进入战斗.

        :return: 进入战斗状态信息，包括['success', 'dock is full'].
        """
        if self.config.chapter != self.timer.port.chapter or self.config.map != self.timer.port.map:
            self._go_map_page()
            self._change_fight_map(self.config.chapter, self.config.map)
            self.timer.port.chapter = self.config.chapter
            self.timer.port.map = self.config.map
        self.timer.wait_images(
            [self.info.map_image, IMG.identify_images['fight_prepare_page']],
            timeout=3,
        )
        self._go_fight_prepare_page()
        move_team(self.timer, self.config.fleet_id)
        if (
            self.config.fleet is not None
            and self.timer.port.fleet[self.config.fleet_id] != self.config.fleet
        ):
            change_ships(self.timer, self.config.fleet_id, self.config.fleet)
            self.timer.port.fleet[self.config.fleet_id] = self.config.fleet[:]

        self.info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.config.repair_mode, self.info.ship_stats)

        return start_march(self.timer)

    def _make_decision(
        self,
        *args,
        **kwargs,
    ):
        state = self.update_state() if 'skip_update' not in kwargs else self.info.state
        if state == ConditionFlag.SL:
            return ConditionFlag.SL

        # 进行 MapLevel 的决策
        if state == 'map_page':
            self.info.fight_history.add_event(
                '自动回港',
                {'position': self.info.node, 'info': '正常'},
            )
            return ConditionFlag.FIGHT_END

        if state == 'fight_condition':
            value = self.config.fight_condition
            self.timer.relative_click(*value.relative_click_position)
            self.info.last_action = value
            self.info.fight_history.add_event(
                '战况选择',
                {'position': self.info.node},
                value,
            )
            return ConditionFlag.FIGHT_CONTINUE

        # 不在白名单之内 SL
        if self.info.node not in self.config.selected_nodes:
            # 可以撤退点撤退
            if state == 'spot_enemy_success':
                self.timer.click(677, 492, delay=0)
                self.info.last_action = 'retreat'
                self.info.fight_history.add_event(
                    '索敌成功',
                    {'position': self.info.node, 'enemies': '不在预设点, 不进行索敌'},
                    '撤退',
                )
                return ConditionFlag.FIGHT_END
            # 不能撤退退游戏
            if state == 'formation':
                self.info.fight_history.add_event(
                    '阵型选择',
                    {'position': self.info.node},
                    'SL',
                )
                return ConditionFlag.SL
            if state == 'fight_period':
                self.info.fight_history.add_event(
                    '进入战斗',
                    {'position': self.info.node},
                    'SL',
                )
                return ConditionFlag.SL

        elif state == 'proceed':
            is_proceed = self.nodes[self.info.node].config.proceed and check_blood(
                self.info.ship_stats,
                self.nodes[self.info.node].config.proceed_stop,
            )

            if is_proceed:
                self.timer.click(325, 350)
                self.info.last_action = 'yes'
                self.info.fight_history.add_event(
                    '继续前进',
                    {'position': self.info.node, 'ship_stats': self.info.ship_stats},
                    '前进',
                )
                return ConditionFlag.FIGHT_CONTINUE
            self.timer.click(615, 350)
            self.info.last_action = 'no'
            self.info.fight_history.add_event(
                '继续前进',
                {'position': self.info.node, 'ship_stats': self.info.ship_stats},
                '回港',
            )
            return ConditionFlag.FIGHT_END

        elif state == 'flagship_severe_damage':
            self.timer.click_image(IMG.fight_image[4], must_click=True, delay=0.25)
            self.info.fight_history.add_event(
                '自动回港',
                {'position': self.info.node, 'info': '旗舰大破'},
            )
            return 'fight end'

        # 进行通用 NodeLevel 决策
        action, fight_stage = self.nodes[self.info.node].make_decision(
            state,
            self.info.last_state,
            self.info.last_action,
            self.info,
        )
        self.info.last_action = action

        return fight_stage

    # ======================== Functions ========================
    def _get_chapter(self) -> int:
        """在出征界面获取当前章节(远征界面也可获取)

        Raises:
            TimeoutError: 无法获取当前章节

        Returns:
            int: 当前章节
        """
        for try_times in range(5):
            time.sleep(0.15 * 2**try_times)
            self.timer.update_screen()
            for i in range(1, len(IMG.chapter_image)):
                if self.timer.image_exist(IMG.chapter_image[i], 0):
                    return i

        raise TimeoutError("can't verify chapter")

    def _move_chapter(self, target_chapter, now_chapter=None):
        """移动地图章节到 target_chapter
        含错误检查

        Args:
            target_chapter (int): 目标
            now_chapter (_type_, optional): 现在的章节. Defaults to None.
        Raise:
            ImageNotFoundErr:如果没有找到章节标志或地图界面标志
        """
        if not self.timer.identify_page('map_page'):
            raise ImageNotFoundErr("not on page 'map_page' now")

        if now_chapter == target_chapter:
            return
        try:
            if now_chapter is None:
                now_chapter = self._get_chapter()
            if self.timer.config.show_chapter_info:
                self.timer.logger.debug('NowChapter:', now_chapter)
            if now_chapter > target_chapter:
                if now_chapter - target_chapter >= 3:
                    now_chapter -= 3
                    self.timer.click(95, 97, delay=0)

                elif now_chapter - target_chapter == 2:
                    now_chapter -= 2
                    self.timer.click(95, 170, delay=0)

                elif now_chapter - target_chapter == 1:
                    now_chapter -= 1
                    self.timer.click(95, 229, delay=0)

            else:
                if now_chapter - target_chapter <= -3:
                    now_chapter += 3
                    self.timer.click(95, 485, delay=0)

                elif now_chapter - target_chapter == -2:
                    now_chapter += 2
                    self.timer.click(95, 416, delay=0)

                elif now_chapter - target_chapter == -1:
                    now_chapter += 1
                    self.timer.click(95, 366, delay=0)

            if not self.timer.wait_image(IMG.chapter_image[now_chapter]):
                raise ImageNotFoundErr(
                    "after 'move chapter' operation but the chapter do not move",
                )

            time.sleep(0.15)
            self._move_chapter(target_chapter, now_chapter)
        except:

            self.logger.warning(
                f'切换章节失败 target_chapter: {target_chapter}   now: {now_chapter}',
            )
            if self.timer.process_bad_network('move_chapter'):
                self._move_chapter(target_chapter)
            else:
                raise ImageNotFoundErr("unknown reason can't find chapter image")

    def _verify_map(self, target_map, chapter, need_screen_shot=True, timeout=0):
        if timeout == 0:
            return self.timer.image_exist(
                IMG.normal_map_image[f'{chapter!s}-'][target_map - 1],
                need_screen_shot,
                confidence=0.85,
            )
        return self.timer.wait_image(
            IMG.normal_map_image[f'{chapter!s}-'][target_map - 1],
            confidence=0.85,
            timeout=timeout,
            gap=0.03,
        )

    def _get_map(self, chapter, need_screen_shot=True) -> int:
        """出征界面获取当前显示地图节点编号
        例如在出征界面显示的地图 2-5,则返回 5

        Returns:
            int: 节点编号
        """
        for try_times in range(5):
            time.sleep(0.15 * 2**try_times)
            if need_screen_shot:
                self.timer.update_screen()

            # 通过+-1来匹配0，1开始的序号
            for map in range(1, MAP_NUM[chapter - 1] + 1):
                if self._verify_map(map, chapter, need_screen_shot=False):
                    return map

        raise TimeoutError("can't verify map")

    def _move_map(self, target_map, chapter, retry_times=2):
        """改变地图节点,不检查是否有该节点
        含网络错误检查
        Args:
            target_map (_type_): 目标节点
            retry_times (int): 重试次数
        """
        if not self.timer.identify_page('map_page'):
            raise ImageNotFoundErr("not on page 'map_page' now")

        now_map = self._get_map(chapter)
        try:
            if self.timer.config.show_chapter_info:
                self.timer.logger.debug('now_map:', now_map)
            if target_map > now_map:
                for i in range(target_map - now_map):
                    self.timer.click(937, 277)
                    # self.timer.swipe(715, 147, 552, 147, duration=0.25)
                    if not self._verify_map(now_map + (i + 1), chapter, timeout=4):
                        raise ImageNotFoundErr(
                            "after 'move map' operation but the chapter do not move",
                        )
                    time.sleep(0.15)
            else:
                for i in range(now_map - target_map):
                    self.timer.click(247, 277)
                    # self.timer.swipe(552, 147, 715, 147, duration=0.25)
                    if not self._verify_map(now_map - (i + 1), chapter, timeout=4):
                        raise ImageNotFoundErr(
                            "after 'move map' operation but the chapter do not move",
                        )
                    time.sleep(0.15)
        except:
            if self.timer.process_bad_network():
                self._move_map(target_map, chapter)
            elif retry_times > 0:
                self.timer.logger.warning(
                    f'切换地图失败, 目标: {target_map}, 当前: {now_map}, 进行重试',
                )
                self._move_map(target_map, chapter, retry_times - 1)
            else:
                self.timer.logger.error(
                    f'切换地图失败, 目标: {target_map}, 当前: {now_map}, 重试失败',
                )
                raise ImageNotFoundErr(
                    "unknown reason can't find number image" + str(target_map),
                )

    def _change_fight_map(self, chapter, map):
        """活动多点战斗必须重写该模块"""
        """ 这个模块负责的工作在战斗结束后如果需要进行重复战斗, 则不会进行 """
        """在地图界面改变战斗地图(通常是为了出征)
        可以处理网络错误
        Args:
            chapter (int): 目标章节
            map (int): 目标节点

        Raises:
            ValueError: 不在地图界面
            ValueError: 不存在的节点
        """
        if self.timer.now_page.name != 'map_page':
            raise ValueError(
                "can't change fight map at page:",
                self.timer.now_page.name,
            )
        if map - 1 not in range(MAP_NUM[chapter - 1]):
            raise ValueError(
                f'map {map!s} not in the list of chapter {chapter!s}',
            )

        self._move_chapter(chapter)
        self._move_map(map, chapter)
        self.info.chapter = self.config.chapter
        self.info.map = self.config.map



import json
import os
import re
import subprocess
import time
from typing import Protocol

import airtest.core.android
import requests
from airtest.core.api import connect_device
from airtest.core.error import AdbError, DeviceConnectionError

from autowsgr.configs import UserConfig
from autowsgr.constants.custom_exceptions import CriticalErr
from autowsgr.types import EmulatorType
from autowsgr.utils.logger import Logger


class OSController(Protocol):
    logger: Logger
    emulator_type: EmulatorType
    emulator_name: str
    emulator_start_cmd: str
    emulator_process_name: str
    dev_name: str  # emulator_name 添加类似 Android:/// 前缀, 自动设置

    def check_network(self) -> bool:
        """检查网络状况

        Returns:
            bool:网络正常返回 True,否则返回 False
        """
        response = requests.get('https://www.moefantasy.com', timeout=5)

        return response.status_code == 200

    def wait_network(self, timeout=1000) -> bool:
        """等待到网络恢复"""
        start_time = time.time()
        while time.time() - start_time <= timeout:
            if self.check_network():
                return True
            time.sleep(10)

        return False

    def is_android_online(self) -> bool: ...

    def connect_android(self) -> airtest.core.android.Android:
        """连接指定安卓设备
        Returns:
            dev: airtest.
        """
        if not self.is_android_online():
            self.restart_android()
            time.sleep(15)
        start_time = time.time()
        while time.time() - start_time <= 30:
            try:
                dev = connect_device(self.dev_name)
                dev.snapshot()
                self.logger.info('Android Connected!')
                return dev
            except (AdbError, DeviceConnectionError):
                self.logger.error('Adb 连接模拟器失败, 正在清除原有连接并重试')
                from airtest.core.android.adb import ADB

                adb = ADB().get_adb_path()
                subprocess.run([adb, 'kill-server'])

        self.logger.error('连接模拟器失败！')
        raise CriticalErr('连接模拟器失败！')

    def kill_android(self) -> None: ...

    def start_android(self) -> None: ...

    def restart_android(self) -> None:
        self.kill_android()
        self.start_android()


class WindowsController(OSController):
    def __init__(
        self,
        config: UserConfig,
        logger: Logger,
    ) -> None:
        self.logger = logger

        self.emulator_type = config.emulator_type
        self.emulator_name = config.emulator_name
        self.emulator_start_cmd = config.emulator_start_cmd
        self.emulator_process_name = config.emulator_process_name
        self.dev_name = f'Android:///{self.emulator_name}'

    def is_android_online(self) -> bool:
        """判断 timer 给定的设备是否在线
        Returns:
            bool: 在线返回 True, 否则返回 False
        """
        match self.emulator_type:
            case EmulatorType.leidian:
                raw_res = self.__ldconsole('isrunning')
                self.logger.debug('EmulatorType status: ' + raw_res)
                return raw_res == 'running'
            case EmulatorType.yunshouji:
                return True
            case _:
                # TODO: 检查是否所有windows版本返回都是中文
                raw_res = subprocess.check_output(
                    f'tasklist /fi "ImageName eq {self.emulator_process_name}',
                ).decode('gbk')
                return 'PID' in raw_res

    def kill_android(self) -> None:
        try:
            match self.emulator_type:
                case EmulatorType.leidian:
                    self.__ldconsole('quit')
                case EmulatorType.yunshouji:
                    self.logger.info('云手机无需关闭')
                case _:
                    subprocess.run(['taskkill', '-f', '-im', self.emulator_process_name])
        except Exception as e:
            raise CriticalErr(f'停止模拟器失败: {e}')

    def start_android(self) -> None:
        try:
            match self.emulator_type:
                case EmulatorType.leidian:
                    self.__ldconsole('launch')
                case EmulatorType.yunshouji:
                    self.logger.info('云手机无需启动')
                case _:
                    os.popen(self.emulator_start_cmd)

            start_time = time.time()
            while not self.is_android_online():
                time.sleep(1)
                if time.time() - start_time > 120:
                    raise TimeoutError('模拟器启动超时！')
        except Exception as e:
            raise CriticalErr(f'模拟器启动失败: {e}')

    def __ldconsole(self, command, command_arg='', global_command=False) -> str:
        """
        使用雷电命令行控制模拟器。

        :param command: 要执行的ldconsole命令。
        :type command: str

        :param command_arg: 命令的附加参数（可选）。
        :type command_arg: str, 可选

        :param global_command: 指示命令是否是全局的（不特定于模拟器实例）。
        :type global_command: bool, 可选

        :return: 雷电命令行执行的输出。
        :rtype: str
        """
        console_dir = os.path.join(os.path.dirname(self.emulator_start_cmd), 'ldconsole.exe')
        emulator_index = (int(re.search(r'\d+', self.emulator_name).group()) - 5554) / 2

        if not global_command:
            cmd = [
                console_dir,
                command,
                '--index',
                str(emulator_index),
                command_arg,
            ]
        else:
            cmd = [console_dir, command_arg]

        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        output, error = process.communicate()

        return (
            output.decode('utf-8', errors='replace')
            if output
            else error.decode('utf-8', errors='replace')
        )


class MacController(OSController):
    def __init__(self, config: UserConfig, logger: Logger) -> None:
        self.logger = logger

        self.emulator_type = config.emulator_type
        self.emulator_name = config.emulator_name
        self.emulator_start_cmd = config.emulator_start_cmd
        self.emulator_process_name = config.emulator_process_name
        self.dev_name = f'Android:///{self.emulator_name}'

        if self.emulator_type == EmulatorType.mumu:
            self.mumu_tool = os.path.join(config.emulator_start_cmd, 'Contents/MacOS/mumutool')
            self.prot = self.emulator_name.split(':')[-1]

    def is_android_online(self) -> bool:
        try:
            subprocess.check_output(f'pgrep -f {self.emulator_process_name}', shell=True)
            if self.emulator_type == EmulatorType.mumu:
                marsh = self.__get_mumu_info()
                return any(self.prot == v.get('adb_port') for v in marsh['return']['results'])
            return True
        except subprocess.CalledProcessError:
            return False

    def kill_android(self) -> None:
        if self.emulator_type == EmulatorType.mumu:
            return
        subprocess.Popen(f'pkill -9 -f {self.emulator_process_name}', shell=True)

    def start_android(self) -> None:
        subprocess.Popen(f'open -a {self.emulator_start_cmd}', shell=True)
        if self.emulator_type == EmulatorType.mumu:
            marsh = self.__get_mumu_info()
            for k, v in enumerate(marsh['return']['results']):
                if self.prot == v.get('adb_port'):
                    cmd = f'{self.mumu_tool} restart {k}'
                    subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,
                    )

    def __get_mumu_info(self) -> dict:
        cmd = f'{self.mumu_tool} info all'
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        output, error = process.communicate()
        tempStr = output.decode()
        try:
            return json.loads(tempStr)
        except Exception as e:
            self.logger.error(f'{cmd} {e}')
        return {}



import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.fight.event.event import Event
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.timer import Timer


NODE_POSITION = (
    (0, 0),
    (0.583, 0.206),
    (0.439, 0.472),
    (0.663, 0.672),
    (0.280, 0.170),
    (0.159, 0.733),
    (0.759, 0.252),
)


class EventFightPlan20250123(Event, NormalFightPlan):
    def __init__(
        self,
        timer: Timer,
        plan_path,
        fleet_id=None,
        event='20250123',
    ) -> None:
        """
        Args:
            fleet_id : 新的舰队参数, 优先级高于 plan 文件, 如果为 None 则使用计划参数.
        """
        if os.path.isabs(plan_path):
            plan_path = plan_path
        else:
            plan_path = timer.plan_tree['event'][event][plan_path]

        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        Event.__init__(self, timer, event)

    def _load_fight_info(self):
        self.info = EventFightInfo20250123(self.timer, self.config.chapter, self.config.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'event', self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """选择并进入战斗地图(chapter-map)"""
        self.change_difficulty(chapter_id)

    def _go_map_page(self):
        self.timer.go_main_page()
        self.timer.click_image(self.event_image[3], timeout=10)

    def _go_fight_prepare_page(self) -> None:
        assert isinstance(self.config.map, int)
        if not self.timer.image_exist(self.event_image[1]):
            self.timer.relative_click(*NODE_POSITION[self.config.map])

        if not self.timer.click_image(self.event_image[1], timeout=10):
            self.timer.logger.warning('进入战斗准备页面失败,重新尝试进入战斗准备页面')
            self.timer.relative_click(*NODE_POSITION[self.config.map])
            self.timer.click_image(self.event_image[1], timeout=10)

        try:
            self.timer.wait_pages('fight_prepare_page', after_wait=0.15)
        except Exception as e:
            self.timer.logger.warning(f'匹配fight_prepare_page失败, 尝试重新匹配, error: {e}')
            self.timer.go_main_page()
            self._go_map_page()
            self._go_fight_prepare_page()


class EventFightInfo20250123(Event, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20250123') -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        Event.__init__(self, timer, event)
        self.map_image = (
            self.common_image['easy']
            + self.common_image['hard']
            + [self.event_image[1]]
            + [self.event_image[2]]
        )
        self.end_page = 'unknown_page'
        self.state2image['map_page'] = [self.map_image, 5]



import time

from autowsgr.constants.ui import Node
from autowsgr.timer.timer import Timer, try_to_get_expedition


class Expedition:
    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        self.is_ready = False
        self.last_check = time.time()

    def update(self, force=False):
        self.timer.update_screen()
        if (isinstance(self.timer.now_page, str) and 'unknown' in self.timer.now_page) or (
            isinstance(self.timer.now_page, Node)
            and self.timer.now_page.name
            not in [
                'expedition_page',
                'map_page',
                'battle_page',
                'exercise_page',
                'decisive_battle_entrance',
            ]
        ):
            if force or time.time() - self.last_check > 1800:
                self.timer.go_main_page()
            if isinstance(self.timer.now_page, Node) and self.timer.now_page.name == 'main_page':
                self.is_ready = self.timer.check_pixel(
                    (933, 454),
                    bgr_color=(45, 89, 255),
                )
        else:
            self.is_ready = self.timer.check_pixel((464, 11), bgr_color=(45, 89, 255))

    def run(self, force=False):
        """检查远征, 如果有未收获远征, 则全部收获并用原队伍继续

        Args:
            force (bool): 是否强制检查
        Returns:
            bool: 是否进行了远征操作
        """
        self.update(force=force)
        if self.is_ready:
            self.timer.goto_game_page('expedition_page')
            try_to_get_expedition(self.timer)
            self.timer.last_expedition_check_time = time.time()
        else:
            self.timer.logger.debug('暂无已完成的远征任务')



"""舰队问答类活动抄这个"""

import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.fight.event.event import Event
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.timer import Timer


NODE_POSITION = (
    None,
    (240, 180),
    (240, 360),
    (480, 180),
    (480, 360),
    (720, 180),
    (720, 360),
)


class EventFightPlan20230613(Event, NormalFightPlan):
    def __init__(
        self,
        timer: Timer,
        plan_path,
        from_alpha=True,
        fleet_id=None,
        event='20230613',
    ) -> None:
        """
        Args:
            fleet_id : 新的舰队参数, 优先级高于 plan 文件, 如果为 None 则使用计划参数.

            from_alpha : 指定入口, 默认为 True 表示从 alpha 进入, 如果为 False 则从 beta 进入, 优先级高于 plan 文件, 如果为 None 则使用计划文件的参数, 如果都没有指定, 默认从 alpha 进入
        """
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        Event.__init__(self, timer, event)

        if from_alpha is not None:
            self.from_alpha = from_alpha

        # 如果未指定入口则默认从 alpha 进入
        if 'from_alpha' not in self.__dict__:
            self.from_alpha = True

    def _load_fight_info(self):
        self.info = EventFightInfo20230613(self.timer, self.chapter, self.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'event', self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """选择并进入战斗地图(chapter-map)"""
        self.change_difficulty(chapter_id)

    def _is_alpha(self):
        return self.timer.check_pixel((795, 321), (254, 148, 36), screen_shot=True)

    def _go_fight_prepare_page(self) -> None:
        if not self.timer.image_exist(self.info.event_image[1], need_screen_shot=0):
            self.timer.click(*NODE_POSITION[self.map])

        # 选择入口
        if self._is_alpha() != self.from_alpha:
            entrance_position = [(820, 377), (820, 321)]
            self.timer.click(*entrance_position[int(self.from_alpha)])

        self.timer.wait_image(self.event_image[1])
        self.timer.click(850, 490)
        self.timer.wait_pages('fight_prepare_page', after_wait=0.15)


class EventFightInfo20230613(Event, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20230613') -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        Event.__init__(self, timer, event)
        self.map_image = (
            self.common_image['easy'] + self.common_image['hard'] + [self.event_image[1]]
        )
        self.end_page = 'unknown_page'
        self.state2image['map_page'] = [self.map_image, 5]



ALL_SHIP_TYPES = [
    'BB',
    'BC',
    'CL',
    'CV',
    'CA',
    'CVL',
    'NAP',
    'DD',
    'SS',
    'CLT',
    'KP',
    'CGBM',
    'AV',
    'AADG',
    'ASDG',
    'BG',
    'CBG',
    'NO',
    'BBV',
    'CAV',
    'SC',
    'SAP',
    'KP',
]
ALL_SHIP_TYPES_CN = [
    '轻巡',
    '驱逐',
    '航母',
    '潜艇',
    '导驱',
    '航战',
    '重巡',
    '防驱',
    '战列',
    '战巡',
    '导驱',
    '导战',
    '大巡',
    '补给',
    '机场',
    '雷巡',
    '装母',
    '轻母',
    '补给',
    '重炮',
    '航巡',
    '导巡',
    '防巡',
]

CG = 'CG'
KP = 'KP'
BB = 'BB'
BC = 'BC'
CL = 'CL'
CV = 'CV'
CA = 'CA'
CVL = 'CVL'
NAP = 'NAP'
DD = 'DD'
SS = 'SS'
CLT = 'CLT'
BM = 'BM'
AV = 'AV'
AADG = 'AADG'
ASDG = 'ASDG'
BG = 'BG'
CBG = 'CBG'
NO = 'NO'
BBV = 'BBV'
CAV = 'CAV'
SC = 'SC'
SAP = 'SAP'

CN_TYPE_TO_EN_TYPE = {
    '轻巡': CL,
    '重巡': CA,
    '雷巡': CLT,
    '航巡': CAV,
    '大巡': CBG,
    '导巡': KP,
    '防巡': CG,
    '航母': CV,
    '轻母': CVL,
    '装母': AV,
    '战列': BB,
    '航战': BBV,
    '战巡': BC,
    '重炮': BM,
    '驱逐': DD,
    '炮潜': SC,
    '潜艇': SS,
    '补给': NAP,
    '导驱': ASDG,
    '防驱': AADG,
    '导战': BG,
    '机场': 'AF',
}

"""舰船类型,参考游戏中显示的字符串,特殊和陌生的命名在下面给出,路基等单位暂时无数据
    NAP:常规补给舰, 胖次期特殊点位刷六个那种,或者说除了掉胖次的之外的所有(包括我方)
    SAP:特殊补给舰, 胖次期掉胖次的蓝色补给舰
    CBG:导弹巡洋舰
    CAV:航巡
    SC:炮潜
    BG:导战
    NO:该位置不存在舰船
"""

ALL_PAGES = {
    'map_page',
    'main_page',
    'decisive_battle_entrance',
    'exercise_page',
    'battle_page',
    'expedition_page',
    'fight_prepare_page',
    'bath_page',
    'choose_repair_page',
    'backyard_page',
    'canteen_page',
    'options_page',
    'build_page',
    'destroy_page',
    'develop_page',
    'discard_page',
    'intensify_page',
    'remake_page',
    'skill_page',
    'mission_page',
    'support_set_page',
    'friend_page',
    'friend_home_page',
}
"""名字说明:
'fight_prepare_page':包含了快速修理,补给,综合属性等选项的界面
'friend_home_page':好友演习界面(从好友界面点进去之后的，好友的提督室)
"""

RESOURCE_NAME = {
    'oil',
    'ammo',
    'steel',
    'aluminum',
    'diamond',
    'quick_repair',
    'quick_build',
    'ship_blueprint',
    'equipment_blueprint',
}

# 船舱容量,装备容量
PORT_FACILITIES = {'ship_limit', 'ship_amount', 'equipment_limit', 'equipment_amount'}

# 维修船坞, 建造位, 开发位

CHAPTER_NODE_COUNT = [0, 5, 6, 4, 4, 5, 4, 5, 5, 4]
# 常规图节点数



import logging
from typing import Any

from airtest.core.settings import Settings as ST  # noqa

from autowsgr.configs import UserConfig
from autowsgr.game.build import BuildManager
from autowsgr.timer import Timer
from autowsgr.utils.io import yaml_to_dict
from autowsgr.utils.logger import Logger
from autowsgr.utils.update import check_for_updates


def start_script(settings_path: str | None = None) -> Timer:
    """启动脚本, 返回一个 Timer 记录器.
    :如果模拟器没有运行, 会尝试启动模拟器,
    :如果游戏没有运行, 会自动启动游戏,
    :如果游戏在后台, 会将游戏转到前台
    Returns:
        Timer: 该模拟器的记录器
    """
    # airtest全局设置
    ST.CVSTRATEGY = ['tpl']

    # config
    config_dict: dict[str, Any] = {} if settings_path is None else yaml_to_dict(settings_path)
    config = UserConfig.from_dict(config_dict)
    config.pprint()

    if config.check_update:
        check_for_updates()

    # logger
    logging.getLogger('airtest').setLevel(logging.ERROR)
    logger = Logger(config.log_dir, config.log_level)
    logger.save_config(config)

    timer = Timer(config, logger)
    timer.port.factory = BuildManager(timer)

    return timer



class ImageNotFoundErr(BaseException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NetworkErr(BaseException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ShipNotFoundErr(BaseException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class CriticalErr(BaseException):
    """严重错误
    通常发生严重错误表明实现有重大问题
    Args:
        BaseException (_type_): _description_
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class LogitException(BaseException):
    def __init__(self, *args: object) -> None:
        super().__init__()(*args)



import ctypes
import os
import sys
from ctypes import POINTER, c_char, c_char_p, c_int32, c_size_t, c_uint8, c_void_p, cast, cdll

import numpy


class ApiDll:
    def __init__(self, path) -> None:
        self.dll = cdll.LoadLibrary(os.path.join(path, f'{sys.platform}_image_autowsgrs.bin'))
        self.dll.locate.argtypes = c_void_p, POINTER(c_int32 * 100)
        self.dll.locate.restype = c_int32
        self.dll.recognize_enemy.argtypes = (c_void_p, c_char_p)
        self.dll.recognize_enemy.restype = c_int32
        self.dll.recognize_map.argtypes = [c_void_p]
        self.dll.recognize_map.restype = c_char

    def _wrap_img_input(self, image: numpy.ndarray) -> c_void_p:
        arr = c_size_t * 4
        width = c_size_t(image.shape[1])
        height = c_size_t(image.shape[0])
        channels = c_size_t(1) if len(image.shape) == 2 else c_size_t(image.shape[2])
        img = image.astype(numpy.uint8)
        pixels_p = POINTER(c_uint8)
        pixels_p = c_size_t(ctypes.cast(img.ctypes.data_as(pixels_p), c_void_p).value)
        return cast(arr(width, height, channels, pixels_p), c_void_p)

    def _wrap_recognize_enemy_input(self, images: list[c_void_p]) -> c_void_p:
        images_arr = c_void_p * images.__len__()
        arr = images_arr()
        for i in range(images.__len__()):
            arr[i] = ctypes.cast(images[i], c_void_p)
        input_arr = c_void_p * (2)
        return input_arr(images.__len__(), cast(arr, c_void_p))

    def locate(self, image):
        image_p = self._wrap_img_input(image)
        len = c_int32 * 100
        res = len()
        result: c_int32 = self.dll.locate(image_p, ctypes.pointer(res))
        ret: list[tuple] = [(res[i * 2], res[i * 2 + 1]) for i in range(result // 2)]
        return ret

    def recognize_enemy(self, images: list[numpy.ndarray]) -> str:
        images_p = [self._wrap_img_input(image) for image in images]
        input = self._wrap_recognize_enemy_input(images_p)
        ret = ctypes.create_string_buffer(b'\0', 100)
        self.dll.recognize_enemy(input, ret)
        return ret.value.decode('ascii')

    def recognize_map(self, image: list[numpy.ndarray]):
        image_p = self._wrap_img_input(image)
        return chr(self.dll.recognize_map(image_p)[0])



import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.fight.event.event import Event
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.timer import Timer


NODE_POSITION = [
    None,
    (661, 134),
    (235, 99),
    (103, 211),
    (848, 369),
    (497, 395),
    (220, 358),
]


class EventFightPlan20231215(Event, NormalFightPlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=None, event='20231215') -> None:
        """
        Args:
            fleet_id : 新的舰队参数, 如果为 None 则使用计划参数.
        """
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        Event.__init__(self, timer, event)

    def _load_fight_info(self):
        self.info = EventFightInfo20231215(self.timer, self.chapter, self.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'event', self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """选择并进入战斗地图(chapter-map)"""
        self.change_difficulty(chapter_id)

    def _go_fight_prepare_page(self) -> None:
        if not self.timer.image_exist(self.info.event_image[1], need_screen_shot=0):
            self.timer.click(*NODE_POSITION[self.map])
        self.timer.wait_image(self.event_image[1])
        self.timer.click(850, 490)
        self.timer.wait_pages('fight_prepare_page', after_wait=0.15)


class EventFightInfo20231215(Event, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20231215') -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        Event.__init__(self, timer, event)
        self.map_image = (
            self.common_image['easy'] + self.common_image['hard'] + [self.event_image[1]]
        )
        self.end_page = 'unknown_page'
        self.state2image['map_page'] = [self.map_image, 5]



import json
import os
from datetime import datetime

from loguru import logger

from autowsgr.configs import UserConfig
from autowsgr.utils.io import save_image


class Logger:
    def __init__(self, log_dir: str, log_level: str) -> None:
        self.log_dir = log_dir
        self.log_level = log_level
        os.makedirs(self.log_dir, exist_ok=True)
        self.reset_level()

    def _configure_logger(self, log_level: str | int) -> None:

        # 自定义格式化器
        def custom_format(record):
            level = record['level'].name

            # 绿色：32，红色：31，黄色：33，蓝色：34，淡蓝色：36，清除：0
            if level == 'INFO':
                return '\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[32m{level:^7}\x1b[0m| {message}\n'
            if level == 'ERROR':
                return '\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[31m{level:^7}\x1b[0m| {message}\n'
            if level == 'DEBUG':
                return '\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[0m{level:^7}| {message}\n'
            if level == 'WARNING':
                return '\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[33m{level:^7}\x1b[0m| {message}\n'
            return '\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[31m{level:^7}| {message}\n'

        # 添加文件日志记录器
        logger.add(
            os.path.join(self.log_dir, 'log_{time}.log'),
            level=log_level,
            colorize=True,
            format='{time:HH:mm:ss} autowsgr {level:^7}| {message}',
            catch=True,
            retention='1 week',
        )

        # 添加控制台日志记录器
        # logger.add(
        #    sys.stdout,
        #    level=log_level,
        #    colorize=True,
        #    format="<green>{time:HH:mm:ss}</green> <level>{level}</level> <cyan>autowsgr:</cyan> <level>{message}</level>",
        #    catch=True,
        # )

        # 添加控制台日志记录器
        logger.add(
            lambda msg: print(msg, end=''),
            level=log_level,
            colorize=True,
            format=custom_format,
            catch=True,
        )

    def save_config(self, config: UserConfig):
        with open(os.path.join(self.log_dir, 'config.json'), 'w', encoding='utf-8') as f:
            json.dump(
                config.asdict(),
                f,
                default=str,
                indent=4,
                ensure_ascii=False,
                sort_keys=True,
            )

    def reset_level(self):
        logger.remove()
        self._configure_logger(self.log_level)

    def debug(self, *args):
        logger.debug(str(args))

    def info(self, string: str) -> None:
        logger.info(string)
        # st.write(string)

    def warning(self, string: str) -> None:
        logger.warning('===================WARNING===================')
        logger.warning(string)
        logger.warning('====================END====================')
        # st.write(string)

    def error(self, string: str) -> None:
        logger.error('===================ERROR===================')
        logger.error(string)
        logger.error('====================END====================')

    def log_image(
        self,
        image,
        name=None,
        ignore_existed_image=False,
    ):
        """向默认数据记录路径记录图片
        Args:
            image: 图片,PIL.Image.Image 格式或者 numpy.ndarray 格式
            name (str): 图片文件名
        """
        if name is None:
            name = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

        if 'png' not in name and 'PNG' not in name:
            name += '.PNG'
        path = os.path.join(self.log_dir, name)

        save_image(
            path=path,
            image=image,
            ignore_existed_image=ignore_existed_image,
        )
        self.info(f'图片已保存至{path}')



import math
import os

import numpy as np
from PIL import Image

from autowsgr.constants.colors import COLORS
from autowsgr.constants.data_roots import OCR_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import (
    AADG,
    ASDG,
    AV,
    BB,
    BBV,
    BC,
    BG,
    BM,
    CA,
    CAV,
    CBG,
    CL,
    CLT,
    CV,
    CVL,
    DD,
    NAP,
    NO,
    RESOURCE_NAME,
    SAP,
    SC,
    SS,
)
from autowsgr.constants.positions import BLOOD_BAR_POSITION, TYPE_SCAN_AREA
from autowsgr.timer import Timer
from autowsgr.types import FormationName
from autowsgr.utils.api_image import crop_image
from autowsgr.utils.io import yaml_to_dict
from autowsgr.utils.math_functions import cal_dis, check_color


class Resources:
    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        self.resources = {}

    def detect_resources(self, name=None):
        timer = self.timer
        if name is not None:
            if name in ('normal', 'oil', 'ammo', 'steel', 'aluminum'):
                self.timer.go_main_page()
                self.detect_resources()
            if name == 'quick_repair':
                self.timer.goto_game_page('choose_repair_page')
                self.detect_resources()
            if name == 'quick_build':
                self.timer.goto_game_page('build_page')
                self.detect_resources()
            if name == 'ship_blueprint':
                self.timer.goto_game_page('build_page')
                self.detect_resources()
            if name == 'equipment_blueprint':
                self.timer.goto_game_page('develop_page')
                self.detect_resources()
        else:
            result = get_resources(timer)
            for key, value in result.items():
                self.resources[key] = value

    def ask_resources(self, name, detect=False):
        """查询资源量(不会从游戏中探查,会根据程序维护的数据返回)
        如果游戏脱离程序监控,可能会不准确，需要先 detect

        Args:
            name (资源名称):
                values:
                    refer to constants.other_constants.RESOURCE_NAME
            detect (bool, optional): 是否从游戏中重新探查(如果否，则使用由程序维护的数据)

        Returns:
            int: 资源量
        """
        if name not in RESOURCE_NAME:
            raise ValueError('Unsupported resource name')
        if detect or name not in self.resources:
            self.detect_resources(name)

        return self.resources.get(name)


POS = yaml_to_dict(os.path.join(OCR_ROOT, 'relative_location.yaml'))


def get_resources(timer: Timer):
    """根据 timer 所处界面获取对应资源数据
    部分 case 会没掉,请重写
    """
    timer.go_main_page()
    timer.update_screen()
    image = timer.screen
    ret = {}
    for key in POS['main_page']['resources']:
        image_crop = crop_image(image, *POS['main_page']['resources'][key])
        try:
            ret[key] = timer.recognize_number(image_crop, 'KM.')[1]
        except:
            # 容错处理，如果监测出来不是数字则出错了
            timer.logger.warning(f'读取{key}资源失败')
    timer.logger.info(ret)
    return ret


def get_loot_and_ship(timer: Timer):
    """获取掉落数据"""
    timer.goto_game_page('map_page')
    timer.update_screen()
    image = timer.screen
    ret = {}
    for key in POS['map_page']:
        image_crop = crop_image(image, *POS['map_page'][key])
        result = timer.recognize_number(image_crop, extra_chars='/', allow_nan=True)
        if result:
            if isinstance(result[1], tuple):
                ret[key], ret[key + '_max'] = result[1]
            else:
                # 如果ocr把"/"识别为"1",则使用下列方法
                try:
                    if key == 'loot':
                        ret[key] = int(str(result[1])[:-3])
                        ret[key + '_max'] = 50
                    if key == 'ship':
                        ret[key] = int(str(result[1])[:-4])
                        ret[key + '_max'] = 500
                except:
                    timer.logger.warning(f'读取{key}数量失败')
        else:
            timer.logger.warning(f'读取{key}数量失败')
    try:
        ship_num = ret.get('ship')
        timer.got_ship_num = ship_num if ship_num is not None else 0
    except:
        timer.logger.warning('赋值给got_ship_num失败')
        timer.got_ship_num = 0

    try:
        timer.got_loot_num = ret.get('loot')
        if timer.got_loot_num is None:
            timer.got_loot_num = 0
        else:
            timer.can_get_loot = True
    except:
        timer.logger.warning('赋值给got_loot_num失败')
        timer.got_loot_num = 0

    timer.logger.info(f'已掉落舰船:{timer.got_ship_num}')
    if timer.can_get_loot:
        timer.logger.info(f'已掉落胖次:{timer.got_loot_num}')
    else:
        timer.logger.info('胖次活动未开启, 无法捞胖次')
    return ret


def get_enemy_condition(timer: Timer, type='exercise', *args, **kwargs):
    """获取敌方舰船类型数据并返回一个字典, 具体图像识别为黑箱, 采用 C++ 实现

    Args:
        type (str, optional): 描述情景. Defaults to 'exercise'.
            'exercise': 演习点击挑战后的一横排舰船

            'fight': 索敌成功后的两列三行

    Returns:
        dict: {[SHIP_TYPE]:[SHIP_AMOUNT]}

        example return: {"CV":1, "BB":3, "DD": 2}
    """

    enemy_type_count = {
        CV: 0,
        BB: 0,
        SS: 0,
        BC: 0,
        NAP: 0,
        DD: 0,
        ASDG: 0,
        AADG: 0,
        CL: 0,
        CA: 0,
        CLT: 0,
        CVL: 0,
        NO: 0,
        'all': 0,
        CAV: 0,
        AV: 0,
        BM: 0,
        SAP: 0,
        BG: 0,
        CBG: 0,
        SC: 0,
        BBV: 0,
        'AP': 0,
    }

    if type == 'exercise':
        type = 0
    if type == 'fight':
        type = 1

    if timer.image_exist(IMG.fight_image[12]):
        # 特殊补给舰
        enemy_type_count[SAP] = 1

    # 处理图像并将参数传递给识别图像的程序
    img = Image.fromarray(timer.screen).convert('L')
    img = img.resize((960, 540))
    imgs = []
    for area in TYPE_SCAN_AREA[type]:
        arr = np.array(img.crop(area))
        imgs.append(arr)

    # 获取并解析结果
    res = timer.ocr_backend.bin.recognize_enemy(imgs).split()

    enemy_type_count['ALL'] = 0
    for x in res:
        enemy_type_count[x] += 1
        if x != NO:
            enemy_type_count['ALL'] += 1
    enemy_type_count[NAP] = enemy_type_count['AP'] - enemy_type_count[SAP]
    count = {k: v for k, v in enemy_type_count.items() if v}
    timer.logger.debug('enemies:' + str(count))
    return count


def get_enemy_formation(timer: Timer) -> str:
    """识别敌方队型, 识别失败返回空字符串"""
    left_bottom = (0.11, 0.15)
    right_top = (0.2, 0.05)
    res = timer.recognize(
        crop_image(timer.screen, left_bottom, right_top),
        allowlist=[*FormationName.get_all_chars(), '阵'],
    )
    return res[1] if res is not None else ''


def detect_ship_stats(timer: Timer, type='prepare', previous=None):
    """检查我方舰船的血量状况(精确到红血黄血绿血)并返回

    Args:
        type (str, optional): 描述在哪个界面检查. .
            'prepare': 战斗准备界面

            'sumup': 单场战斗结算界面
    Returns:
        list: 表示血量状态, 1-index

        example: [-1, 0, 0, 1, 1, 2, -1] 表示 1-2 号位绿血 3-4 号位中破, 5 号位大破, 6 号位不存在

    """
    # Todo: 检测是否满血/触发中保, 精确到数值的检测, 战斗结算时检测不依赖先前信息

    timer.update_screen()
    result = [-1, 0, 0, 0, 0, 0, 0]
    for i in range(1, 7):
        if type == 'prepare':
            pixel = timer.get_pixel(*BLOOD_BAR_POSITION[0][i])
            result[i] = check_color(pixel, COLORS.BLOOD_COLORS[0])
            if result[i] in [3, 2]:
                result[i] = 2
            elif result[i] == 0:
                result[i] = 0
            elif result[i] == 4:
                result[i] = -1
            else:
                result[i] = 1
        elif type == 'sumup':
            if previous and previous[i] == -1:
                result[i] = -1
                continue
            pixel = timer.get_pixel(*BLOOD_BAR_POSITION[1][i])
            result[i] = check_color(pixel, COLORS.BLOOD_COLORS[1])
            if result[i] == 4:
                result[i] = -1
    return result


def detect_ship_type(timer: Timer):
    """ToDo
    在出征准备界面读取我方所有舰船类型并返回该列表
    """


def get_exercise_stats(timer: Timer, robot=None) -> list[None] | None:
    """检查演习界面, 第 position 个位置,是否为可挑战状态, 强制要求屏幕中只有四个目标

    Args:
        position (_type_): 编号从屏幕中的位置编起, 如果滑动显示了第 2-5 个敌人, 那么第二个敌人编号为 1

    Returns:
        bool: 如果可挑战, 返回 True , 否则为 False, 1-index
    """
    timer.update_screen()
    up = timer.check_pixel((933, 59), (177, 171, 176), distance=60)
    down = timer.check_pixel((933, 489), (177, 171, 176), distance=60)
    assert not (up and down)

    result: list = [
        None,
    ]
    if not up and not down:
        timer.swipe(800, 200, 800, 400)  # 上滑
        timer.update_screen()
        up = True
    if up:
        for position in range(1, 5):
            result.append(  # noqa: PERF401
                math.sqrt(
                    cal_dis(
                        timer.get_pixel(770, position * 110 - 10),
                        COLORS.CHALLENGE_BLUE,
                    ),
                )
                <= 50,
            )
        timer.swipe(800, 400, 800, 200)  # 下滑
        timer.update_screen()
        result.append(
            math.sqrt(
                cal_dis(timer.get_pixel(770, 4 * 110 - 10), COLORS.CHALLENGE_BLUE),
            )
            <= 50,
        )
        return result
    if down:
        for position in range(1, 5):
            result.append(  # noqa: PERF401
                math.sqrt(
                    cal_dis(
                        timer.get_pixel(770, position * 110 - 10),
                        COLORS.CHALLENGE_BLUE,
                    ),
                )
                <= 50,
            )
        if robot is not None:
            result.insert(1, robot)
        else:
            timer.swipe(800, 200, 800, 400)  # 上滑
            timer.update_screen()
            result.insert(
                1,
                math.sqrt(
                    cal_dis(timer.get_pixel(770, 4 * 110 - 10), COLORS.CHALLENGE_BLUE),
                )
                <= 50,
            )

            timer.swipe(800, 400, 800, 200)  # 下滑

        return result
    return None


def check_support_stats(timer: Timer):
    """在出征准备界面检查是否开启了战役支援(有开始出征按钮的界面)

    Returns:
        bool: 先判断是否为灰色，如果为灰色则返回True，然后判断是否开启，如果开启则返回True，否则返回False
    """
    timer.update_screen()
    pixel = timer.get_pixel(623, 75)
    d1 = cal_dis(pixel, COLORS.SUPPORT_ENABLE)  # 支援启用的黄色
    d2 = cal_dis(pixel, COLORS.SUPPORT_DISABLE)  # 支援禁用的蓝色
    d3 = cal_dis(pixel, COLORS.SUPPORT_ENLESS)  # 支援次数用尽的灰色
    if d1 > d3 and d2 > d3:
        timer.logger.info('战役支援次数已用尽')
        return True
    return d1 < d2



import os
from typing import Protocol

import cv2
import numpy as np

from autowsgr.configs import UserConfig
from autowsgr.constants.data_roots import BIN_ROOT
from autowsgr.timer.backends.api_dll import ApiDll
from autowsgr.utils.io import cv_imread
from autowsgr.utils.logger import Logger


def edit_distance(word1, word2) -> int:
    """
    解题思路，动态规划
        步骤1：将word1与word2前拼接上空格，方便为空时的操作
        步骤2：初始化dp第一个元素，接着初始化dp的第一行与第一列
        步骤3：可通过画表(如: excel里)找到状态转移的规律，填充剩下的dp格子即可
    :return: dp[-1][-1], 返回最后操作的结果
    """
    m, n = len(word1), len(word2)
    if m == 0 and n == 0:
        return 0
    word1, word2 = (
        ' ' + word1,
        ' ' + word2,
    )  # 非常必要的操作，不添加空格话，在Word为空时会比较麻烦
    dp = [[0 for _ in range(n + 1)] for _ in range(m + 1)]
    dp[0][0] = 0  # 初始化dp[0][0] = 0，因为空格对空格不需要任何操作，即0步
    for i in range(1, n + 1):  # 第一行初始化
        dp[0][i] = i
    for i in range(1, m + 1):  # 第一列初始化
        dp[i][0] = i
    for i in range(1, m + 1):  # 逐个填充剩余的dp格子
        for j in range(1, n + 1):
            if word1[i] == word2[j]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = min(dp[i - 1][j - 1], dp[i - 1][j], dp[i][j - 1]) + 1
    return dp[-1][-1]


def find_lcseque(s1, s2):
    """求两个字符串的LCS"""
    m = [[0 for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    d: list[list[str | None]] = [[None for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    for p1 in range(len(s1)):
        for p2 in range(len(s2)):
            if s1[p1] == s2[p2]:
                m[p1 + 1][p2 + 1] = m[p1][p2] + 1
                d[p1 + 1][p2 + 1] = 'ok'
            elif m[p1 + 1][p2] > m[p1][p2 + 1]:
                m[p1 + 1][p2 + 1] = m[p1 + 1][p2]
                d[p1 + 1][p2 + 1] = 'left'
            else:
                m[p1 + 1][p2 + 1] = m[p1][p2 + 1]
                d[p1 + 1][p2 + 1] = 'up'
    (p1, p2) = (len(s1), len(s2))
    s = []
    while m[p1][p2]:  # 不为None时
        c = d[p1][p2]
        if c == 'ok':  # 匹配成功，插入该字符，并向左上角找下一个
            s.append(s1[p1 - 1])
            p1 -= 1
            p2 -= 1
        if c == 'left':  # 根据标记，向左找下一个
            p2 -= 1
        if c == 'up':  # 根据标记，向上找下一个
            p1 -= 1
    s.reverse()
    return ''.join(s)


class OCRBackend(Protocol):
    config: UserConfig
    logger: Logger
    WORD_REPLACE: dict[
        str,
        str,
    ]  # 记录中文ocr识别的错误用于替换。主要针对词表缺失的情况，会导致稳定的识别为另一个字
    bin: ApiDll

    def __init_subclass__(cls) -> None:
        cls.bin = ApiDll(os.path.join(BIN_ROOT))

    def read_text(
        self,
        img,
        allowlist: list[str] | None | str = None,
        sort: str = 'left-to-right',
        **kwargs,
    ):
        """识别文字的具体实现，返回字符串格式识别结果"""
        raise NotImplementedError

    def resize_image_proportionally(self, image, scale_factor):
        """
        等比例扩大图片。

        参数:
        image -- NumPy数组格式的图片
        scale_factor -- 扩大的倍数，例如2表示扩大到原来的两倍

        返回:
        resized_image -- 扩大后的图片
        """
        # 获取原始图片的尺寸
        original_height, original_width = image.shape[:2]

        # 计算新的尺寸
        new_width = int(original_width * scale_factor)
        new_height = int(original_height * scale_factor)

        # 使用cv2.resize函数等比例扩大图片
        return cv2.resize(
            image,
            (new_width, new_height),
            interpolation=cv2.INTER_LINEAR,
        )

    def recognize(
        self,
        img,
        allowlist: list[str] | None | str = None,
        candidates: list[str] | None = None,
        multiple=False,
        allow_nan=False,
        rgb_select=None,
        tolerance=30,
        scale_factor=2,
        **kwargs,
    ):
        """识别任意字符串, 该函数为最底层封装

        Args:
            scale_factor (int, optional): 图标缩放倍率. Defaults to 2.
        Returns:
            list((int, int), str, float): 中心位置, 字串识别结果, 置信度
        """

        def pre_process_rgb(img, rgb_select=None, tolerance=30):
            # 如果没有提供rgb_select，直接返回原始图像
            if rgb_select is None:
                return img

            # 将BGR图像转换为RGB格式
            img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

            rgb_select_normalized = np.array(rgb_select)
            mask = np.all(np.abs(img_rgb - rgb_select_normalized) <= tolerance, axis=-1)

            # 使用掩码将匹配的像素保留，其他像素设置为255
            result_img = img_rgb.copy()
            result_img[mask] = 0
            result_img[~mask] = 255

            # 将处理后的图像转换回BGR格式
            return cv2.cvtColor(result_img, cv2.COLOR_RGB2BGR)

        def split_image(img: np.ndarray):
            def is_key_line(line: np.ndarray):
                return line.min() != 255
                # for i in range(1, line.shape[0], 8):
                #     if any(val != 255 for val in line[i]):
                #         return True
                # return False

            bias = []
            imgs = []
            last_array = np.ndarray([0, img.shape[1], 3], dtype=np.uint8)
            for i in range(img.shape[0]):
                if (not is_key_line(img[i])) or i == img.shape[0] - 1:
                    if last_array.shape[0] > 8:
                        imgs.append(last_array)
                        bias.append(i - last_array.shape[0])
                    last_array = np.ndarray([0, img.shape[1], 3], dtype=np.uint8)
                    # else:
                else:
                    last_array = np.append(last_array, [img[i]], 0)
            return imgs, bias

        def post_process_text(t):
            for k, v in self.WORD_REPLACE.items():
                t = t.replace(k, v)
            res, name = 'dsjfagiahsdifhaoisd', t
            if candidates is None:
                return name
            for _name in candidates:
                if any((_name.find(char) != -1) for char in name):
                    dis1 = edit_distance(_name, name) / (len(find_lcseque(_name, name)) + 1)
                    dis2 = edit_distance(res, name) / (len(find_lcseque(res, name)) + 1)
                    if dis1 < dis2:
                        res = _name
            return res

        img = pre_process_rgb(img, rgb_select, tolerance)
        if type(img) is str:
            img = cv2.imdecode(
                np.frombuffer(cv_imread(img), np.uint8),
                cv2.IMREAD_COLOR,
            )
        img = self.resize_image_proportionally(img, scale_factor)
        imgs, bias = split_image(img)
        results = []
        for img, bia in zip(imgs, bias, strict=False):
            _results = self.read_text(img, allowlist, **kwargs)
            _results = [x for x in _results if x[1] != '']  # 去除空匹配
            _results = [(t[0], post_process_text(t[1]), t[2]) for t in _results]
            for result in _results:
                results.append(  # noqa: PERF401
                    (
                        [
                            (result[0][0]) // scale_factor,
                            (result[0][1] + bia) // scale_factor,
                        ],
                        result[1],
                        result[2],
                    ),
                )
        if self.config.show_ocr_info:
            self.logger.debug(f'修正OCR结果: {results}')

        if allow_nan and not results:
            return None

        if multiple:
            return results
        if not results:
            results = ['Unknown']
        return results[0]

    def recognize_number(
        self,
        img,
        extra_chars='',
        multiple=False,
        allow_nan=False,
        **kwargs,
    ):
        """识别数字"""

        def process_number(t: str):
            # 今日胖次、掉落; 决战升级经验等
            if '/' in t:
                nums = t.split('/')
                assert len(nums) == 2
                return process_number(nums[0]), process_number(nums[1])

            # 决战，费用是f"x{cost}"格式
            t = t.lstrip('xXKk')
            # 战后经验值 f"Lv.{exp}"格式
            t = t.lstrip('Lv.')
            # 建造资源有前导0
            if t != '0':
                t = t.lstrip('0')

            # 资源可以是K/M结尾
            if t.endswith(('K', 'k')):
                return eval(t[:-1]) * 1000
            if t.endswith('M'):
                return eval(t[:-1]) * 1000000

            # 未识别到数字, 返回-1
            if t == '':
                return -1
            return eval(t)

        results = self.recognize(
            img,
            allowlist='0123456789' + extra_chars,
            multiple=True,
            **kwargs,
        )
        results = [(t[0], process_number(t[1]), t[2]) for t in results]
        if self.config.show_ocr_info:
            self.logger.debug(f'数字解析结果：{results}')

        if allow_nan and not results:
            return None

        if multiple:
            return results
        if len(results) != 1:
            self.logger.warning(f'OCR识别数字失败: {results}')
            results = []
        return results[0]

    def recognize_ship(self, image, candidates, **kwargs):
        """传入一张图片,返回舰船信息,包括名字和舰船型号"""
        if isinstance(image, str):
            image_path = os.path.abspath(image)
            img = cv2.imread(image_path)
        else:
            img = image
        location = self.bin.locate(img)
        ret = []
        for i in range(len(location)):
            res = self.recognize(
                img[location[i][0] - 1 : location[i][1] + 1],
                multiple=True,
                candidates=candidates,
                **kwargs,
            )
            # 因为 recognize 的参数 allow_nan 为 False，所以 res 不会为 None
            assert res is not None
            # 因为 recognize 的参数 multiple 为 True，所以 res 不会为 str
            assert not isinstance(res, str)
            for j in range(len(res)):
                res[j][0][1] = res[j][0][1] + location[i][0] - 1
            ret += res

        return ret

    # def recognize_time(self, img, format="%H:%M:%S"):
    #     """识别时间"""
    #     text = self.recognize(img, allowlist="0123456789:").replace(" ", "")
    #     return str2time(text, format)


class EasyocrBackend(OCRBackend):
    def __init__(self, config: UserConfig, logger: Logger) -> None:
        self.config = config
        self.logger = logger
        self.WORD_REPLACE = {
            '鲍鱼': '鲃鱼',
            '鲴鱼': '鲃鱼',
        }
        import easyocr

        self.reader = easyocr.Reader(['ch_sim', 'en'])

    def read_text(
        self,
        img,
        allowlist: list[str] | None | str = None,
        sort='left-to-right',
        # TODO：以下参数可能需要调整，以获得最好OCR性能
        min_size=7,
        text_threshold=0.25,
        low_text=0.3,
        **kwargs,
    ):
        """识别文字的具体实现，返回字符串格式识别结果"""

        def get_center(pos1, pos2):
            x1, y1 = pos1
            x2, y2 = pos2
            return (x1 + x2) / 2, (y1 + y2) / 2

        results = self.reader.readtext(
            img,
            allowlist=allowlist,
            min_size=min_size,
            text_threshold=text_threshold,
            low_text=low_text,
            **kwargs,
        )
        results = [(get_center(r[0][0], r[0][2]), r[1], r[2]) for r in results]

        if sort == 'left-to-right':
            results = sorted(results, key=lambda x: x[0][0])
        elif sort == 'top-to-bottom':
            results = sorted(results, key=lambda x: x[0][1])
        else:
            raise ValueError(f'Invalid sort method: {sort}')

        if self.config.show_ocr_info:
            self.logger.debug(f'原始OCR结果: {results}')
        return results


class PaddleOCRBackend(OCRBackend):

    def __init__(self, config: UserConfig, logger: Logger) -> None:
        self.config = config
        self.logger = logger
        self.WORD_REPLACE = {
            '鲍鱼': '鲃鱼',
        }

        # TODO:后期单独训练模型，提高识别准确率，暂时使用现成的模型
        from paddleocr import PaddleOCR

        self.reader = PaddleOCR(
            use_angle_cls=True,
            use_gpu=True,
            show_log=False,
            lang='ch',
        )  # need to run only once to download and load model into memory

    def read_text(self, img, allowlist, **kwargs):

        def get_center(pos1, pos2):
            x1, y1 = pos1
            x2, y2 = pos2
            return (x1 + x2) / 2, (y1 + y2) / 2

        results = self.reader.ocr(img, cls=False, **kwargs)
        results = [] if results == [None] else results[0]
        results = [(get_center(r[0][1], r[0][3]), r[1][0], r[1][1]) for r in results]
        if self.config.show_ocr_info:
            self.logger.debug(f'原始OCR结果: {results}')
        return results



from autowsgr.configs import BattleConfig
from autowsgr.constants.image_templates import IMG
from autowsgr.fight.common import DecisionBlock, FightInfo, FightPlan, start_march
from autowsgr.game.game_operation import get_ship, quick_repair
from autowsgr.game.get_game_info import detect_ship_stats
from autowsgr.timer import Timer
from autowsgr.types import ConditionFlag
from autowsgr.utils.io import yaml_to_dict


"""
战役模块/单点战斗模板
"""


class BattleInfo(FightInfo):
    def __init__(self, timer: Timer) -> None:
        super().__init__(timer)

        self.end_page = 'battle_page'

        self.successor_states = {
            'proceed': ['spot_enemy_success', 'formation', 'fight_period'],
            'spot_enemy_success': {
                'retreat': ['battle_page'],
                'fight': ['formation', 'fight_period'],
            },
            'formation': ['fight_period'],
            'fight_period': ['night', 'result'],
            'night': {
                'yes': ['result'],
                'no': [['result', 7]],
            },
            'night_fight_period': ['result'],
            'result': ['battle_page'],  # 两页战果
        }

        self.state2image = {
            'proceed': [IMG.fight_image[5], 7.5],
            'spot_enemy_success': [IMG.fight_image[2], 15],
            'formation': [IMG.fight_image[1], 15, 0.8],
            'fight_period': [IMG.symbol_image[4], 7.5],
            'night': [IMG.fight_image[6], 150],
            'result': [IMG.fight_image[16], 75],
            'battle_page': [
                IMG.identify_images.battle_page,
                7.5,
            ],
        }

        self.after_match_delay = {
            'night': 1.75,
            'get_ship': 1,
        }

    def reset(self):
        self.fight_history.reset()
        self.last_state = ''
        self.last_action = ''
        self.state = 'proceed'

    def _before_match(self):
        # 点击加速
        if self.state in ['proceed']:
            self.timer.click(380, 520, delay=0, enable_subprocess=True)
        self.timer.update_screen()

    def _after_match(self):
        if self.state == 'get_ship':
            get_ship(self.timer)
        super()._after_match()


class BattlePlan(FightPlan):
    def __init__(self, timer, plan_path: str | None = None, plan_args: dict | None = None) -> None:
        super().__init__(timer)
        # 加载计划配置
        file_plan_args = (
            yaml_to_dict(self.timer.plan_tree['battle'][plan_path]) if plan_path else {}
        )
        file_plan_args.update(plan_args or {})
        plan_args = file_plan_args
        self.config = BattleConfig.from_dict(plan_args)

        # 加载节点配置
        node_args = plan_args.get('node_args', {})
        self.node = DecisionBlock(timer, node_args)
        self.info = BattleInfo(timer)

    def _go_fight_prepare_page(self):
        self.timer.goto_game_page('battle_page')
        now_hard = self.timer.wait_images([IMG.fight_image[9], IMG.fight_image[15]])
        hard = self.config.map > 5
        if now_hard != hard:
            self.timer.click(800, 80, delay=1)

    def _enter_fight(self) -> ConditionFlag:
        self._go_fight_prepare_page()
        self.timer.click(180 * ((self.config.map - 1) % 5 + 1), 200)
        self.timer.wait_pages('fight_prepare_page', after_wait=0.15)
        self.info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.config.repair_mode, ship_stats=self.info.ship_stats)
        try:
            return start_march(self.timer)
        except TimeoutError:
            self.logger.warning(
                '由于进入战斗超时跳过了一次战役, 请检查战役队伍中是否有舰船正在远征',
            )
            return ConditionFlag.SKIP_FIGHT

    def _make_decision(self, *args, **kwargs) -> ConditionFlag:
        state = self.update_state() if 'skip_update' not in kwargs else self.info.state
        if state == ConditionFlag.SL:
            return ConditionFlag.SL
        if self.info.state == 'battle_page':
            return ConditionFlag.FIGHT_END

        # 进行通用 NodeLevel 决策
        action, fight_stage = self.node.make_decision(
            self.info.state,
            self.info.last_state,
            self.info.last_action,
            self.info,
        )
        self.info.last_action = action
        return fight_stage



import easyocr

from autowsgr.utils.io import cv_imread
from autowsgr.utils.operator import unzip_element


ch_reader = None


def load_ch_reader():
    global ch_reader
    ch_reader = easyocr.Reader(['ch_sim'], gpu=True)


def get_allow(names):
    char_set = set()
    for name in unzip_element(names):
        for char in name:
            char_set.add(char)
    res = ''
    for char in char_set:
        res += char
    return res


def recognize(image, char_list=None, min_size=7, text_threshold=0.55, low_text=0.3):
    if ch_reader is None:
        load_ch_reader()
    if isinstance(image, str):
        image = cv_imread(image)
    assert ch_reader is not None
    return ch_reader.readtext(
        image,
        allowlist=char_list,
        min_size=min_size,
        text_threshold=text_threshold,
        low_text=low_text,
    )


if __name__ == '__main__':
    pass



import os
import re
import subprocess
import time

import inquirer
import pkg_resources
import requests
from packaging.version import parse


update_source = [
    inquirer.List(
        'source',
        message='Please choose the source to update',
        choices=['清华源(推荐)', '北京外国语', 'PyPI'],
    ),
]


def check_for_updates():
    print('Checking for updates...')
    # 获取本地autowsgr版本号
    local_version = get_local_version()
    assert local_version is not None, 'Failed to get the local version'

    # 发送 GET 请求获取库的元数据信息
    response = requests.get('https://mirrors.pku.edu.cn/pypi/web/json/autowsgr', timeout=5)
    data = response.json()

    # 提取最新版本号
    latest_version = data['info']['version']
    # 比较版本号
    if parse(local_version) < parse(latest_version):
        update_questions = [
            inquirer.List(
                'source',
                message=f'New version {latest_version} is available.Your version is {local_version}. Do you want to update?',
                choices=['Yes', 'No'],
            ),
        ]
        result = get_user_choice(update_questions)
        if result == 'Yes':
            # 选择使用哪个源更新,输出按钮回车选择
            choice = get_user_choice(update_source)
            update_library(choice)
            recent_updates = get_recent_updates_from_pypi(latest_version)
            print('更新内容:\n' + recent_updates)

            print('更新完成，稍后将自动退出，请重新启动脚本')
            time.sleep(5)
            os._exit(0)  # 更新成功后退出脚本
    else:
        print('You are using the latest version of the library.')


def get_local_version():
    # 使用pkg_resources获取本地库的版本号
    try:
        return pkg_resources.get_distribution('autowsgr').version
    except Exception as e:
        print(f'Failed to get the local version.Error: {e}')
        return None


def get_user_choice(questions):
    answers = inquirer.prompt(questions)
    assert answers is not None, 'User did not make a choice'
    return answers['source']


def update_library(choice='PyPI'):
    choice_list = {
        'PyPI': [
            'pip',
            'install',
            '--upgrade',
            '--index-url',
            'https://pypi.org/simple',
            'autowsgr',
        ],
        '北京外国语': [
            'pip',
            'install',
            '--index-url',
            'https://mirrors.bfsu.edu.cn/pypi/web/simple/',
            '--upgrade',
            'autowsgr',
        ],
        '清华源(推荐)': [
            'pip',
            'install',
            '--index-url',
            'https://pypi.tuna.tsinghua.edu.cn/simple',
            '--upgrade',
            'autowsgr',
        ],
    }
    subprocess.run(choice_list[choice])


def get_recent_updates_from_pypi(latest_version):

    url = f'https://pypi.org/project/autowsgr/{latest_version}/#description'
    response = requests.get(url, timeout=5)

    if response.status_code == 200:
        readme_content = response.text
        updates_section = re.search(
            r'<h2>近期更新</h2>(.*?)</ul>',
            readme_content,
            re.DOTALL,
        )

        if updates_section:
            updates = updates_section.group(1).strip()
            # 提取所有 <li> 标签中的内容
            updates_list = re.findall(r'<li>(.*?)</li>', updates, re.DOTALL)
            # 合并为一个字符串，每行前面加上一个 ·
            return '\n'.join(
                [f'· {re.sub(r"<.*?>", "", update).strip()}' for update in updates_list],
            )
        return '未找到近期更新部分。'
    return f'无法获取更新内容，状态码: {response.status_code}'


# if __name__ == "__main__":
#     check_for_updates()



from functools import partial

from airtest.core.cv import Template

from autowsgr.constants.data_roots import IMG_ROOT
from autowsgr.utils.io import create_namespace


class MyTemplate(Template):
    def __radd__(self, other) -> list:
        if isinstance(other, list):
            return [*other, self]  # 添加到列表开头
        return NotImplemented

    def __add__(self, other) -> list:
        if isinstance(other, list):
            return [self, *other]  # 添加到列表末尾
        return NotImplemented


IMG = create_namespace(
    IMG_ROOT,
    partial(MyTemplate, threshold=0.9, resolution=(960, 540)),
)



import numpy as np
from scipy.spatial import distance


def cal_dis(p1: list[int] | tuple[int, ...], p2: list[int] | tuple[int, ...]):
    """计算两个点的欧几里得距离的平方
    Args:
        p1 : 第一个点的坐标
        p2 : 第二个点的坐标

    Raises:
        TypeError: 如果两个点的维度不同则抛出该异常

    Returns:
        float: 表示这两点之间欧几里得距离的平方和
    """
    if len(p1) != len(p2):
        raise ValueError('Dimensions do not equal')
    return distance.euclidean(p1, p2) ** 2


def get_nearest(position, points):
    """计算一个点距离points的哪个点距离更近
    Args:
        position : 第一个点的坐标，(x,y)
        points : 待计算的坐标，是一个列表

    Returns:
        距离最近的点在list中位于第几个位置
    """
    if not True:
        raise ValueError('no color template')
    result = 0 if points[0] is not None else 1
    for i in range(1, len(points)):
        if cal_dis(position, points[i]) < cal_dis(position, points[result]):
            result = i
    return result


def check_color(col, color_list):
    """给定一个颜色和待选颜色列表,返回最相近的颜色的下标 (RGB空间欧几里得距离比较)

    Args:
        col (_type_): 目标颜色(RGB)
        color_list (_type_): 待选颜色列表(RGB)

    Returns:
        int: 表示和最接近的颜色在 color_list 中的下标 0-based
    Raise:
        ValueError: 如果没有待选颜色列表则抛该异常
    """
    return get_nearest(col, color_list)


def matrix_to_str(matrix: np.ndarray):
    """将一个矩阵转化为字符串,格式为:第一行两个正整数 n,m 表示行数和列数,接下来 n 行每行 m 列

    For Example:
        >>> print(matrix_to_str([[1, 2, 3], [2, 3, 4]]))
        2 3
        1 2 3
        2 3 4

    Args:
        matrix (numpy.ndarray): 矩阵

    Raises:
        ValueError: 如果 matrix 参数不是二维的

    Returns:
        str: 结果字符串
    """
    shape = matrix.shape
    if len(shape) != 2:
        raise ValueError('matrix must be a 2D ndarray')
    res = str(len(matrix)) + ' ' + str(len(matrix[0])) + '\n'
    for i in range(len(matrix)):
        x = matrix[i]
        for j in range(len(x)):
            y = x[j]
            res += str(y) + ' '
        res += '\n'
    return res



"""获取 UI 树"""


class SwitchMethod:
    def __init__(self, fun_list) -> None:
        self.operates = fun_list

    def operate(self):
        return list(self.operates)

    def print(self):
        for operation in self.operates:
            print(operation, end=' ')


class Node:
    """节点
    保存 UI 树的节点
    """

    def __init__(self, name: str, id: int) -> None:
        self.id: int = id
        self.name: str = name
        self.father_edge: Edge | None = None
        self.father: Node | None = None
        self.depth: int = 0
        self.edges: list[Edge] = []

    def set_father(self, father):
        self.father = father
        self.father_edge = self.find_edge(father)

    def add_edge(self, edge):
        self.edges.append(edge)

    def find_edges(self, v):
        return [edge for edge in self.edges if edge.v is v]

    def find_edge(self, v):
        from random import choice

        edges = self.find_edges(v)
        return choice(edges) if len(edges) != 0 else None

    def print(self):
        print('节点名:', self.name, '节点编号:', self.id)
        print('父节点:', self.father)
        print('节点连边:')
        for edge in self.edges:
            edge.print()

    def __str__(self) -> str:
        return self.name


class Edge:
    """图的边
    保存 UI 图的边
    """

    def __init__(
        self,
        operate_fun: SwitchMethod,
        u: Node,
        v: Node,
        other_dst=None,
        extra_op: SwitchMethod | None = None,
    ) -> None:
        self.operate_fun = operate_fun
        self.u = u
        self.v = v
        self.other_dst = other_dst
        self.extra_op = extra_op

    def operate(self):
        return self.operate_fun.operate()

    def print(
        self,
    ):
        print('起点:', self.u.name, '终点:', self.v.name)


class UI:
    """UI树(拓扑学概念)"""

    def __init__(self) -> None:
        self.nodes: dict[str, Node] = {}
        self.page_count = 0
        self.is_normal_fight_prepare = False
        self._build_ui_tree()

    def get_node_by_name(self, name):
        return self.nodes.get(name)

    def page_exist(self, page):
        return page in self.nodes.values()

    def find_path(self, start: Node, end: Node):
        lca = self._lca(start, end)
        path1, path2 = [], []

        while start != lca:
            path1.append(start)
            # 因为 lca 存在，所以 start 在不是 lca 的时候一定存在 father
            assert start.father is not None
            start = start.father
        while end != lca:
            path2.append(end)
            # 因为 lca 存在，所以 end 在不是 lca 的时候一定存在 father
            assert end.father is not None
            end = end.father
        path2.reverse()
        path: list[Node] = [*path1, lca, *path2]
        result, i = [], 0
        while i < len(path):
            node = path[i]
            result.append(node)
            if node.find_edge(path[-1]) is not None:
                return [*result, path[-1]]
            for j in range(i + 2, len(path)):
                dst = path[j]
                if node.find_edge(dst) is not None:
                    i = j - 1
            i += 1
        return result

    def print(self):
        for node in self.nodes.values():
            node.print()

    def _build_ui_tree(self):
        main_page = self._construct_node('main_page', None)
        (
            map_page,
            exercise_page,
            expedition_page,
            battle_page,
            decisive_battle_entrance,
        ) = self._construct_integrative_pages(
            main_page,
            names=[
                'map_page',
                'exercise_page',
                'expedition_page',
                'battle_page',
                'decisive_battle_entrance',
            ],
            click_positions=[(163, 25), (287, 25), (417, 25), (544, 25), (661, 25)],
            common_edges=[{'pos': (30, 30), 'dst': main_page}],
        )

        options_page = self._construct_node('options_page', main_page)
        (
            build_page,
            destroy_page,
            develop_page,
            discard_page,
        ) = self._construct_integrative_pages(
            options_page,
            names=[
                'build_page',
                'destroy_page',
                'develop_page',
                'discard_page',
            ],
            click_positions=[(163, 25), (287, 25), (417, 25), (544, 25)],
            common_edges=[{'pos': (30, 30), 'dst': options_page}],
        )
        intensify_page, remake_page, skill_page = self._construct_integrative_pages(
            options_page,
            names=[
                'intensify_page',
                'remake_page',
                'skill_page',
            ],
            click_positions=[(163, 25), (287, 25), (417, 25)],
            common_edges=[{'pos': (30, 30), 'dst': options_page}],
        )

        fight_prepare_page = self._construct_node('fight_prepare_page', map_page)
        backyard_page = self._construct_node('backyard_page', main_page)
        bath_page = self._construct_node('bath_page', backyard_page)
        choose_repair_page = self._construct_node('choose_repair_page', bath_page)
        canteen_page = self._construct_node('canteen_page', backyard_page)
        mission_page = self._construct_node('mission_page', main_page)
        support_set_page = self._construct_node('support_set_page', main_page)
        friend_page = self._construct_node('friend_page', options_page)

        self._add_edge(
            main_page,
            map_page,
            self._construct_clicks_method([(900, 480, 1, 0)]),
            expedition_page,
        )
        self._add_edge(
            main_page,
            mission_page,
            self._construct_clicks_method([(656, 480, 1, 0)]),
        )
        self._add_edge(
            main_page,
            backyard_page,
            self._construct_clicks_method([(45, 80, 1, 0)]),
        )
        self._add_edge(
            main_page,
            support_set_page,
            self._construct_clicks_method([(50, 135, 1, 1), (200, 300, 1, 1)]),
        )
        self._add_edge(
            main_page,
            options_page,
            self._construct_clicks_method([(42, 484, 1, 0)]),
        )

        self._add_edge(
            map_page,
            fight_prepare_page,
            self._construct_clicks_method([(600, 300, 1, 0)]),
        )

        self._add_edge(
            fight_prepare_page,
            map_page,
            self._construct_clicks_method([(33, 30, 1, 0)]),
        )
        self._add_edge(
            fight_prepare_page,
            bath_page,
            self._construct_clicks_method([(840, 20, 1, 0)]),
        )

        self._add_edge(
            options_page,
            build_page,
            self._construct_clicks_method([(150, 200, 1, 1.25), (360, 200, 1, 0)]),
            develop_page,
        )
        self._add_edge(
            options_page,
            remake_page,
            self._construct_clicks_method([(150, 270, 1, 1.25), (360, 270, 1, 0)]),
        )
        self._add_edge(
            options_page,
            friend_page,
            self._construct_clicks_method([(150, 410, 1, 0)]),
        )
        self._add_edge(
            options_page,
            main_page,
            self._construct_clicks_method([(36, 500, 1, 0)]),
        )

        self._add_edge(
            backyard_page,
            canteen_page,
            self._construct_clicks_method([(700, 400, 1, 0)]),
        )
        self._add_edge(
            backyard_page,
            bath_page,
            self._construct_clicks_method([(300, 200, 1, 0)]),
        )
        self._add_edge(
            backyard_page,
            main_page,
            self._construct_clicks_method([(50, 30, 1, 0)]),
        )

        self._add_edge(
            bath_page,
            main_page,
            self._construct_clicks_method([(120, 30, 1, 0)]),
        )
        self._add_edge(
            bath_page,
            choose_repair_page,
            self._construct_clicks_method([(900, 30, 1, 0)]),
        )

        self._add_edge(
            choose_repair_page,
            bath_page,
            self._construct_clicks_method([(916, 45, 1, 0)]),
        )

        self._add_edge(
            canteen_page,
            main_page,
            self._construct_clicks_method([(120, 30, 1, 0)]),
        )
        self._add_edge(
            canteen_page,
            backyard_page,
            self._construct_clicks_method([(50, 30, 1, 0)]),
        )

        self._add_edge(
            mission_page,
            main_page,
            self._construct_clicks_method([(30, 30, 1, 0)]),
        )

        self._add_edge(
            support_set_page,
            main_page,
            self._construct_clicks_method([(30, 30, 1, 0.5), (50, 30, 1, 0.5)]),
        )

        self._add_edge(
            friend_page,
            options_page,
            self._construct_clicks_method([(30, 30, 1, 0)]),
        )

        self._dfs(main_page)

    def _add_node(self, node: Node):
        self.nodes[node.name] = node

    def _lca(self, u: Node, v: Node) -> Node:
        if v.depth > u.depth:
            return self._lca(v, u)
        if u == v:
            return v
        # 因为 u 的 depth 大于 v 的 depth，所以 u.father 一定存在
        assert u.father is not None
        if u.depth == v.depth:
            # 因为 u.depth == v.depth 并且 u != v 即 u 和 v 不是根节点，所以 v.father 一定存在
            assert v.father is not None
            return self._lca(u.father, v.father)
        return self._lca(u.father, v)

    def _dfs(self, u: Node):
        for edge in u.edges:
            son = edge.v
            if son == u.father or son.father != u:
                continue

            son.depth = u.depth + 1
            self._dfs(son)

    def _list_walk_path(self, start: Node, end: Node):
        path = self.find_path(start, end)
        for node in path:
            print(node, end='->')

    def _construct_node(self, name: str, father: Node | None):
        self.page_count += 1
        node = Node(name, self.page_count)
        node.set_father(father)
        self._add_node(node)
        return node

    def _construct_clicks_method(self, click_position_args):
        operations = [['click', operation] for operation in click_position_args]

        return SwitchMethod(operations)

    def _add_edge(self, u: Node, v: Node, method: SwitchMethod, other_dst=None):
        edge = Edge(method, u, v, other_dst=other_dst)
        u.add_edge(edge)

    def _construct_integrative_pages(
        self,
        father,
        click_positions=None,
        names=None,
        common_edges=None,
    ) -> list[Node]:
        if click_positions is None:
            click_positions = []
        if names is None:
            names = []
        if common_edges is None:
            common_edges = []

        assert len(click_positions) == len(names)
        first_node = self._construct_node(names[0], father)
        nodes = [first_node] + [self._construct_node(name, first_node) for name in names[1:]]
        for i, node in enumerate(nodes):
            for j, click_position in enumerate(click_positions):
                if i == j:
                    continue
                self._add_edge(
                    node,
                    nodes[j],
                    self._construct_clicks_method([click_position]),
                )

            for edge in common_edges:
                dst = edge.get('dst')
                x, y = edge.get('pos')
                self._add_edge(node, dst, self._construct_clicks_method([(x, y)]))
        return nodes


WSGR_UI = UI()



from collections.abc import Iterable
from typing import Any


def remove_0_value_from_dict(d: dict) -> dict:
    return {k: v for k, v in d.items() if v is not None and v != 0}


def unzip_element(o: object) -> list[object] | Any | list:
    if not isinstance(o, Iterable):
        return [o]
    res = []
    for value in o:
        if isinstance(value, list | set):
            res += unzip_element(value)
        else:
            res.append(value)
    return res


def unorder_equal(o1: Iterable, o2: Iterable, skip=None) -> bool:
    if skip is None:
        skip = [None]
    return (set(o1) - set(skip)) == (set(o2) - set(skip))
